# Copyright 2021-2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

import numpy as np
import numba

from netket.errors import concrete_or_error, NumbaOperatorGetConnDuringTracingError


from .compile_helpers import pack_internals
from .base import LocalOperatorBase

if TYPE_CHECKING:
    from .jax import LocalOperatorJax


class LocalOperator(LocalOperatorBase):
    """A custom local operator. This is a sum of an arbitrary number of operators
    acting locally on a limited set of k quantum numbers (i.e. k-local,
    in the quantum information sense).
    """

    __module__ = "netket.operator"

    def _setup(self, force: bool = False):
        """Analyze the operator strings and precompute arrays for get_conn inference"""
        if force or not self._initialized:
            data = pack_internals(
                self.hilbert,
                self._operators_dict,
                self.constant,
                self.dtype,
                self.mel_cutoff,
            )

            self._acting_on = data["acting_on"]
            self._acting_size = data["acting_size"]
            self._diag_mels = data["diag_mels"]
            self._mels = data["mels"]
            self._x_prime = data["x_prime"]
            self._n_conns = data["n_conns"]
            self._basis = data["basis"]
            self._nonzero_diagonal = data["nonzero_diagonal"]
            self._max_conn_size = data["max_conn_size"]
            self._initialized = True

    def get_conn_flattened(self, x, sections, pad=False):
        r"""Finds the connected elements of the Operator. Starting
        from a given quantum number x, it finds all other quantum numbers x' such
        that the matrix element :math:`O(x,x')` is different from zero. In general there
        will be several different connected states x' satisfying this
        condition, and they are denoted here :math:`x'(k)`, for :math:`k=0,1...N_{\mathrm{connected}}`.

        This is a batched version, where x is a matrix of shape (batch_size,hilbert.size).

        Args:
            x (matrix): A matrix of shape (batch_size,hilbert.size) containing
                        the batch of quantum numbers x.
            sections (array): An array of size (batch_size) useful to unflatten
                        the output of this function.
                        See numpy.split for the meaning of sections.
            pad (bool): Whether to use zero-valued matrix elements in order to return all equal sections.

        Returns:
            matrix: The connected states x', flattened together in a single matrix.
            array: An array containing the matrix elements :math:`O(x,x')` associated to each x'.

        """
        self._setup()

        x_ids = self.hilbert.states_to_local_indices(x)

        x_ids = concrete_or_error(
            np.asarray,
            x_ids,
            NumbaOperatorGetConnDuringTracingError,
            self,
        )

        xp_ids, mels = self._get_conn_flattened_kernel(
            x_ids,
            sections,
            self._basis,
            self._constant,
            self._diag_mels,
            self._n_conns,
            self._mels,
            self._x_prime,
            self._acting_on,
            self._acting_size,
            self._nonzero_diagonal,
            pad,
        )

        xp = self.hilbert.local_indices_to_states(xp_ids, dtype=x.dtype)
        return xp, mels

    @staticmethod
    @numba.jit(nopython=True)
    def _get_conn_flattened_kernel(
        x,
        sections,
        basis,
        constant,
        diag_mels,
        n_conns,
        all_mels,
        all_x_prime,
        acting_on,
        acting_size,
        nonzero_diagonal,
        pad=False,
    ):
        batch_size = x.shape[0]
        n_sites = x.shape[1]
        dtype = all_mels.dtype

        # TODO remove this line when numba 0.53 is dropped 0.54 is minimum version
        # workaround a bug in Numba arising when NUMBA_BOUNDSCHECK=1
        constant = constant.item()

        assert sections.shape[0] == batch_size

        n_operators = n_conns.shape[0]
        # array to store the row index
        xs_n = np.empty((batch_size, n_operators), dtype=np.intp)

        tot_conn = 0
        max_conn = 0

        for b in range(batch_size):
            # diagonal element
            conn_b = 1 if nonzero_diagonal else 0

            # counting the off-diagonal elements
            for i in range(n_operators):
                acting_size_i = acting_size[i]
                # compute the number (row index) from the local states
                # (here we do the inverse of _number_to_state from
                # compile_helpers.py, so this is essentially _state_to_number)
                xs_n[b, i] = 0
                x_b = x[b]
                x_i = x_b[acting_on[i, :acting_size_i]]
                # iterate over sites the current operator is acting on
                for k in range(acting_size_i):
                    # compute
                    xs_n[b, i] += x_i[acting_size_i - k - 1] * basis[i, k]

                # sum the number of off-diagonal connected elements
                conn_b += n_conns[i, xs_n[b, i]]

            tot_conn += conn_b
            sections[b] = tot_conn

            if pad:
                max_conn = max(conn_b, max_conn)

        if pad:
            tot_conn = batch_size * max_conn

        x_prime = np.empty((tot_conn, n_sites), dtype=x.dtype)
        mels = np.empty(tot_conn, dtype=dtype)

        c = 0
        for b in range(batch_size):
            c_diag = c
            x_batch = x[b]

            if nonzero_diagonal:
                mels[c_diag] = constant
                x_prime[c_diag] = np.copy(x_batch)
                c += 1

            for i in range(n_operators):
                if nonzero_diagonal:
                    mels[c_diag] += diag_mels[i, xs_n[b, i]]

                # get the number of connected elements for the current operator
                # at the rows index corresponding to the state of x at the sites
                # the operator is acting on
                n_conn_i = n_conns[i, xs_n[b, i]]

                if n_conn_i > 0:
                    sites = acting_on[i]
                    acting_size_i = acting_size[i]

                    for cc in range(n_conn_i):  # iterate over compressed nonzero cols
                        # get the nonzero mels of the current row
                        mels[c + cc] = all_mels[i, xs_n[b, i], cc]
                        x_prime[c + cc] = np.copy(x_batch)
                        # set the changed local states of the sites the operator
                        # is acting on
                        # it is stored in all_x_prime, where we select the row
                        for k in range(acting_size_i):
                            x_prime[c + cc, sites[k]] = all_x_prime[
                                i, xs_n[b, i], cc, k
                            ]
                    c += n_conn_i

            if pad:
                delta_conn = max_conn - (c - c_diag)
                mels[c : c + delta_conn].fill(0)
                x_prime[c : c + delta_conn, :] = np.copy(x_batch)
                c += delta_conn
                sections[b] = c

        return x_prime, mels

    def get_conn_filtered(self, x, sections, filters):
        r"""Finds the connected elements of the Operator using only a subset of operators. Starting
        from a given quantum number x, it finds all other quantum numbers x' such
        that the matrix element :math:`O(x,x')` is different from zero. In general there
        will be several different connected states x' satisfying this
        condition, and they are denoted here :math:`x'(k)`, for :math:`k=0,1...N_{\mathrm{connected}}`.

        This is a batched version, where x is a matrix of shape (batch_size,hilbert.size).

        Args:
            x (matrix): A matrix of shape (batch_size,hilbert.size) containing
                        the batch of quantum numbers x.
            sections (array): An array of size (batch_size) useful to unflatten
                        the output of this function.
                        See numpy.split for the meaning of sections.
            filters (array): Only operators op(filters[i]) are used to find the connected elements of
                        x[i].

        Returns:
            matrix: The connected states x', flattened together in a single matrix.
            array: An array containing the matrix elements :math:`O(x,x')` associated to each x'.

        """
        self._setup()

        x_ids = self.hilbert.states_to_local_indices(x)
        x_ids = concrete_or_error(
            np.asarray,
            x_ids,
            NumbaOperatorGetConnDuringTracingError,
            self,
        )

        xp_ids, mels = self._get_conn_filtered_kernel(
            x_ids,
            sections,
            self._basis,
            self._constant,
            self._diag_mels,
            self._n_conns,
            self._mels,
            self._x_prime,
            self._acting_on,
            self._acting_size,
            filters,
        )
        xp = self.hilbert.local_indices_to_states(xp_ids, dtype=x.dtype)
        return xp, mels

    @staticmethod
    @numba.jit(nopython=True)
    def _get_conn_filtered_kernel(
        x,
        sections,
        basis,
        constant,
        diag_mels,
        n_conns,
        all_mels,
        all_x_prime,
        acting_on,
        acting_size,
        filters,
    ):
        batch_size = x.shape[0]
        n_sites = x.shape[1]
        dtype = all_mels.dtype

        assert filters.shape[0] == batch_size and sections.shape[0] == batch_size

        # TODO remove this line when numba 0.53 is dropped 0.54 is minimum version
        # workaround a bug in Numba arising when NUMBA_BOUNDSCHECK=1
        constant = constant.item()

        n_operators = n_conns.shape[0]
        xs_n = np.empty((batch_size, n_operators), dtype=np.intp)

        tot_conn = 0

        for b in range(batch_size):
            # diagonal element
            tot_conn += 1

            # counting the off-diagonal elements
            i = filters[b]

            assert i < n_operators and i >= 0
            acting_size_i = acting_size[i]

            xs_n[b, i] = 0
            x_b = x[b]
            x_i = x_b[acting_on[i, :acting_size_i]]
            for k in range(acting_size_i):
                xs_n[b, i] += x_i[acting_size_i - k - 1] * basis[i, k]

            tot_conn += n_conns[i, xs_n[b, i]]
            sections[b] = tot_conn

        x_prime = np.empty((tot_conn, n_sites))
        mels = np.empty(tot_conn, dtype=dtype)

        c = 0
        for b in range(batch_size):
            c_diag = c
            mels[c_diag] = constant
            x_batch = x[b]
            x_prime[c_diag] = np.copy(x_batch)
            c += 1

            i = filters[b]
            # Diagonal part
            mels[c_diag] += diag_mels[i, xs_n[b, i]]
            n_conn_i = n_conns[i, xs_n[b, i]]

            if n_conn_i > 0:
                sites = acting_on[i]
                acting_size_i = acting_size[i]

                for cc in range(n_conn_i):
                    mels[c + cc] = all_mels[i, xs_n[b, i], cc]
                    x_prime[c + cc] = np.copy(x_batch)

                    for k in range(acting_size_i):
                        x_prime[c + cc, sites[k]] = all_x_prime[i, xs_n[b, i], cc, k]
                c += n_conn_i

        return x_prime, mels

    def to_jax_operator(self) -> "LocalOperatorJax":  # noqa: F821
        """
        Returns the jax-compatible version of this operator, which is an
        instance of :class:`netket.operator.LocalOperatorJax`.
        """
        from .jax import LocalOperatorJax

        return LocalOperatorJax(
            self.hilbert,
            self.operators,
            self.acting_on,
            self.constant,
            dtype=self.dtype,
            mel_cutoff=self.mel_cutoff,
        )



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common utilities to define fields and properties of netket dataclasses
and Pytrees.
"""
from functools import partial

import dataclasses
from dataclasses import MISSING

import jax

from .pytree_serialization_sharding import ShardedFieldSpec


def _cache_name(property_name):
    return "__" + property_name + "_cache"


def _raw_cache_name(cache_name):
    # removes leading __ and ending _cache
    return cache_name[2:-6]


## Our stuff
class _Uninitialized:
    """
    Sentinel value used to signal uninitialized values
    """

    def __repr__(self):
        return "Uninitialized"


Uninitialized = _Uninitialized()

jax.tree_util.register_pytree_node(
    _Uninitialized, lambda x: ((), Uninitialized), lambda *args: Uninitialized
)


def field(
    pytree_node: bool = True,
    pytree_ignore: bool = False,
    serialize: bool | None = None,
    serialize_name: str | None = None,
    cache: bool = False,
    sharded: bool | ShardedFieldSpec = False,
    **kwargs,
):
    """Mark a field of a dataclass or PyTree to be:

    Args:
        pytree_node: a leaf node in the pytree representation of this dataclass.
            If False this must be hashable.
        pytree_ignore: If True this field is ignored by the pytree metadata and will be
            excluded from the pytree flattening/unflattening process. This means the field
            will not appear in the flattened representation when calling `jax.tree_util.tree_flatten()`
            or similar pytree operations. This is useful for caches, temporary data, or other
            fields that should not be passed forward during pytree transformations.
            When True, pytree_node must be False.
        serialize: If True the node is included in the serialization.
            In general you should not specify this. (Defaults to value of pytree_node).
        serialize_name: If specified, it's the name under which this attribute is serialized.
            This can be used to change the runtime attribute name, but maintain some
            other name in the serialisation format.
        cache: If True this node is a cache and will be reset every time
            fields are modified.
        sharded: a boolan or specification object specifying whether this entry is sharded.
            Defaults to False. If True, a MPI-compatible sharding along axis 0 is assumed.
    """
    if serialize is None:
        serialize = pytree_node
    if sharded is True:
        sharded = ShardedFieldSpec()
    if pytree_ignore and pytree_node:
        raise ValueError(
            "Cannot set both `pytree_node=True` and `pytree_ignore=True`. Pytreenode must be False if pytree_ignore is True."
        )

    metadata = {
        "pytree_node": pytree_node,
        "ignore": pytree_ignore,
        "serialize": serialize,
        "cache": cache,
        "sharded": sharded,
    }
    if serialize_name is not None:
        metadata["serialize_name"] = serialize_name
    return dataclasses.field(
        metadata=metadata,
        **kwargs,
    )


def static_field(**kwargs):
    return field(pytree_node=False, **kwargs)


class CachedProperty:
    """Sentinel attribute wrapper to signal that a method is a property
    but must be cached.
    """

    def __init__(self, method, pytree_node=False, pytree_ignore=False):
        self.name = method.__name__
        self.cache_name = _cache_name(self.name)
        self.method = method
        self.pytree_node = pytree_node
        self.pytree_ignore = pytree_ignore
        self.type = method.__annotations__.get("return", MISSING)
        self.doc = method.__doc__

        if self.type is MISSING:
            raise TypeError(
                f"Cached property {method} requires a return type annotation."
            )

    def __get__(self, obj, objtype=None):
        val = getattr(obj, self.cache_name, Uninitialized)
        if val is Uninitialized:
            val = self.method(obj)
            setattr(obj, self.cache_name, val)
        return val

    def __repr__(self):
        return (
            f"CachedProperty(name={self.name}, "
            f"type={self.type}, pytree_node={self.pytree_node})"
        )


def property_cached(fun=None, pytree_node=False, pytree_ignore=False):
    """Decorator to make the method behave as a property but cache the resulting value and
    clears it upon replace.

    Args:
        pytree_node: a leaf node in the pytree representation of this dataclass.
            If False this must be hashable
    """
    if pytree_ignore and pytree_node:
        raise ValueError(
            "Cannot set both `pytree_node=True` and `pytree_ignore=True`. Pytreenode must be False if pytree_ignore is True."
        )

    if fun is None:
        return partial(
            property_cached, pytree_node=pytree_node, pytree_ignore=pytree_ignore
        )

    return CachedProperty(fun, pytree_node=pytree_node, pytree_ignore=pytree_ignore)



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains two classes used to index into hilbert spaces
inheriting from `HomogeneousHilbert`.

Those classes provide an informal API that can be used to extend or override
netket's indexing logic, which is particularly relevant when working with
constrained Hilbert spaces.

----------------------------------------------------------------------------
    This is not part of NetKet's public API and may change at any moment!
----------------------------------------------------------------------------

An hilbert indexing class should respect the `HilbertIndex` protocol
defined in the file `base.py`.

"""

from .base import HilbertIndex, is_indexable, max_states
from .unconstrained import LookupTableHilbertIndex
from .uniform_tensor import UniformTensorProductHilbertIndex
from .constrained_generic import ConstrainedHilbertIndex, optimalConstrainedHilbertindex
from .constrained_sum import SumConstrainedHilbertIndex
from .constrained_sum_partitions import SumOnPartitionConstrainedHilbertIndex


from netket.utils import _hide_submodules

_hide_submodules(__name__)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .continuous_hilbert import ContinuousHilbert
from netket.experimental.geometry import Cell


class Particle(ContinuousHilbert):
    r"""Hilbert space derived from ContinuousHilbert defining N particles
    in continuous space with or without periodic boundary conditions."""

    def __init__(
        self,
        N: int | tuple[int, ...],
        *,
        geometry: Cell,
    ) -> None:
        """
        Constructs new ``Particles`` given specifications
        of the continuous space they are defined in.

        Args:
            N: Number of particles. If int all have the same spin. If Tuple the entry indicates how many particles
                there are with a certain spin-projection.
            geometry: Instance of :class:`~netket.experimental.geometry.Cell` defining the geometry.
        """

        if not hasattr(N, "__len__"):
            N = (N,)

        if not isinstance(geometry, Cell):
            raise TypeError(
                "`geometry` must be an instance of `netket.experimental.geometry.Cell`."
            )

        self._N = sum(N)
        self._n_per_spin = N
        self._geometry = geometry

        super().__init__(geometry.extent)

    @property
    def size(self) -> int:
        return self._N * len(self.domain)

    @property
    def n_particles(self) -> int:
        r"""The number of particles"""
        return self._N

    @property
    def n_per_spin(self) -> tuple[int, ...]:
        r"""Gives the number of particles in a specific spin
        projection.

        The length of this tuple indicates the total spin whereas
        the position in the tuple indicates the spin projection.

        Example: (10,5,3) describes 18 particles of total spin 1
        where 10 of those have spin-projection -1, 5 have
        spin-projection 0 and 3 have spin-projection 1.
        """
        return self._n_per_spin

    @property
    def geometry(self) -> Cell:
        """Geometry of the continuous space."""
        return self._geometry

    @property
    def _attrs(self):
        return (self._N, self.geometry)

    def __repr__(self):
        return f"Particle(N={self.n_particles}, d={len(self.domain)})"



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ["FermionOperator2nd", "FermionOperator2ndJax"]

from .pyscf import from_pyscf_molecule
from ._particle_number_conserving_fermionic import (
    ParticleNumberConservingFermioperator2nd,
    ParticleNumberAndSpinConservingFermioperator2nd,
    FermiHubbardJax,
)


from . import fermion

from netket.operator import FermionOperator2nd as _deprecated_FermionOperator2nd
from netket.operator import FermionOperator2ndJax as _deprecated_FermionOperator2ndJax

_deprecations = {
    # June 2024, NetKet 3.13
    "FermionOperator2nd": (
        "netket.experimental.operator.FermionOperator2nd is deprecated: use "
        "netket.operator.FermionOperator2nd (netket >= 3.13)",
        _deprecated_FermionOperator2nd,
    ),
    "FermionOperator2ndJax": (
        "netket.experimental.operator.FermionOperator2ndJax is deprecated: use "
        "netket.operator.FermionOperator2ndJax (netket >= 3.13)",
        _deprecated_FermionOperator2ndJax,
    ),
}


from netket.utils import _auto_export
from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr

__getattr__ = _deprecation_getattr(__name__, _deprecations)
_auto_export(__name__)

del _deprecation_getattr, _auto_export



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np


from netket.operator import DiscreteOperator
from netket.hilbert import SpinOrbitalFermions
from netket.utils.optional_deps import import_optional_dependency
from netket.operator import FermionOperator2nd, FermionOperator2ndJax
from ._particle_number_conserving_fermionic import (
    ParticleNumberConservingFermioperator2nd,
    ParticleNumberAndSpinConservingFermioperator2nd,
)

from ._pyscf_utils import arrays_to_terms, TV_from_pyscf_molecule


def operator_from_arrays(
    const,
    tij_sparse,
    vijkl_sparse,
    n_electrons,
    hi=None,
    cls=FermionOperator2nd,
    term_conj2=(1, 0),
    term_conj4=(1, 0, 1, 0),
):
    """
    H = const + Σ tᵢⱼ cᵢ†cⱼ + Σ vᵢⱼₖₗ cᵢ†cⱼcₖ†cₗ
    use term_conj4=(1,1,0,0) if you want cᵢ†cⱼ†cₖcₗ instead.
    use a zero matrix/tensor of the correct shape for tij and/or vijkl if you only want to use one of them
    """
    if hi is None:
        if isinstance(n_electrons, tuple):
            assert len(n_electrons) == 2
            hi = SpinOrbitalFermions(
                n_orbitals=tij_sparse.shape[0] // 2,
                s=1 / 2,
                n_fermions_per_spin=n_electrons,
            )
        else:
            hi = SpinOrbitalFermions(
                n_orbitals=tij_sparse.shape[0], n_fermions_per_spin=n_electrons
            )
    terms, weights, constant = arrays_to_terms(
        const, tij_sparse, vijkl_sparse, term_conj2=term_conj2, term_conj4=term_conj4
    )
    return cls(hi, terms, weights, constant)


def from_pyscf_molecule(
    molecule,  # type: pyscf.gto.mole.Mole  # noqa: F821
    mo_coeff: np.ndarray | None = None,
    *,
    cutoff: float = 1e-11,
    implementation: DiscreteOperator = ParticleNumberAndSpinConservingFermioperator2nd,
) -> DiscreteOperator:
    r"""
    Construct a netket operator encoding the electronic hamiltonian of a pyscf
    molecule in a chosen orbital basis.

    Example:
        Constructs the hamiltonian for a Li-H molecule, using the `sto-3g` basis
        and the default Hartree-Fock molecular orbitals.

        >>> from pyscf import gto, scf, fci
        >>> import netket as nk; import netket.experimental as nkx
        >>>
        >>> bond_length = 1.5109
        >>> geometry = [('Li', (0., 0., -bond_length/2)), ('H', (0., 0., bond_length/2))]
        >>> mol = gto.M(atom=geometry, basis='STO-3G')
        >>>
        >>> mf = scf.RHF(mol).run(verbose=0)  # doctest:+ELLIPSIS
        >>> E_hf = sum(mf.scf_summary.values())
        >>>
        >>> E_fci = fci.FCI(mf).kernel()[0]
        >>>
        >>> ha = nkx.operator.from_pyscf_molecule(mol)  # doctest:+ELLIPSIS
        >>> E0 = float(nk.exact.lanczos_ed(ha))
        >>> print(f"{E0 = :.5f}, {E_fci = :.5f}")
        E0 = -7.88253, E_fci = -7.88253

    Example:
        Constructs the hamiltonian for a Li-H molecule, using the `sto-3g` basis
        and the Boys orbitals using :class:`~pyscf.lo.Boys`.

        >>> from pyscf import gto, scf, lo
        >>> import netket as nk; import netket.experimental as nkx
        >>>
        >>> bond_length = 1.5109
        >>> geometry = [('Li', (0., 0., -bond_length/2)), ('H', (0., 0., bond_length/2))]
        >>> mol = gto.M(atom=geometry, basis='STO-3G')
        >>>
        >>> # compute the boys orbitals
        >>> mf = scf.RHF(mol).run(verbose=0)  # doctest:+ELLIPSIS
        >>> mo_coeff = lo.Boys(mol).kernel(mf.mo_coeff)
        >>> # use the boys orbitals to construct the netket hamiltonian
        >>> ha = nkx.operator.from_pyscf_molecule(mol, mo_coeff=mo_coeff)

    Args:
        molecule: The pyscf :class:`~pyscf.gto.mole.Mole` object describing the
            Hamiltonian
        mo_coeff: The molecular orbital coefficients determining the
            linear combination of atomic orbitals to produce the
            molecular orbitals. If unspecified this defaults to
            the hartree fock orbitals computed using :class:`~pyscf.scf.HF`.
        cutoff: Ignores all matrix elements in the `V` and `T` matrix that have
            magnitude less than this value. Defaults to :math:`10^{-11}`
        implementation: The particular implementation to use for the operator.
            Different fermionic operator implementation might have different
            performances. Defaults to
            :class:`netket.experimental.operator.ParticleNumberAndSpinConservingFermioperator2nd` (this might
            change in the future).

    Returns:
        A netket second quantised operator that encodes the electronic hamiltonian.
    """
    # If unspecified, use HF molecular orbitals
    if mo_coeff is None:
        pyscf = import_optional_dependency("pyscf", descr="pyscf_molecule_to_arrays")

        mf = pyscf.scf.HF(molecule).run(verbose=0)
        mo_coeff = mf.mo_coeff

    if implementation in [FermionOperator2ndJax, FermionOperator2ndJax]:
        E_nuc, Tij, Vijkl = TV_from_pyscf_molecule(molecule, mo_coeff, cutoff=cutoff)

        ha = operator_from_arrays(
            E_nuc,
            Tij,
            0.5 * Vijkl,
            molecule.nelec,
            term_conj4=(1, 1, 0, 0),
            cls=implementation,
        )
        # TODO maybe run setup and set _max_conn_size here estimating it analytially
        return ha
    elif implementation in [
        ParticleNumberConservingFermioperator2nd,
        ParticleNumberAndSpinConservingFermioperator2nd,
    ]:
        return implementation.from_pyscf_molecule(molecule, mo_coeff)
    else:
        raise ValueError("unknown implementation")



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
from functools import partial

import jax
import jax.numpy as jnp

from flax import linen as nn

from netket import config
from netket.utils.types import Array, PyTree, PRNGKeyT

# Necessary for the type annotation to work
if config.netket_sphinx_build:
    from netket import sampler

from .base import MetropolisRule


class MultipleRules(MetropolisRule):
    r"""A Metropolis sampling rule that can be used to pick a rule from a list of rules
    with a given probability.

    Each `rule[i]` will be selected with a probability `probabilities[i]`.
    """

    rules: tuple[MetropolisRule, ...]
    """List of rules to be selected from."""
    probabilities: jax.Array
    """Corresponding list of probabilities with which every rule can be
    picked."""

    def __init__(
        self, rules: tuple[MetropolisRule, ...], probabilities: Array
    ) -> MetropolisRule:
        r"""A Metropolis sampling rule that can be used to pick a rule from a list of rules
        with a given probability.

        Each `rule[i]` will be selected with a probability `probabilities[i]`.

        Args:
            rules: A list of rules, one for each subspace of the tensor hilbert space.
            probabilities: A list of probabilities, one for each rule.
        """
        probabilities = jnp.asarray(probabilities)

        if not jnp.allclose(jnp.sum(probabilities), 1.0):
            raise ValueError(
                "The probabilities must sum to 1, but they sum to "
                f"{jnp.sum(probabilities)}."
            )

        if not isinstance(rules, tuple | list) or not all(
            isinstance(r, MetropolisRule) for r in rules
        ):
            raise TypeError(
                "The first argument (rules) must be a tuple of `MetropolisRule` "
                f"rules, but you have passed {type(rules)}."
            )

        if len(probabilities) != len(rules):
            raise ValueError(
                "Length mismatch between the probabilities and the rules: probabilities "
                f"has length {len(probabilities)} , rules has length {len(rules)}."
            )

        self.rules = rules
        self.probabilities = probabilities

    def init_state(
        self,
        sampler: "sampler.MetropolisSampler",  # noqa: F821
        machine: nn.Module,
        params: PyTree,
        key: PRNGKeyT,
    ) -> Any | None:
        N = len(self.probabilities)
        keys = jax.random.split(key, N)
        return tuple(
            self.rules[i].init_state(sampler, machine, params, keys[i])
            for i in range(N)
        )

    def reset(
        self,
        sampler: "sampler.MetropolisSampler",  # noqa: F821
        machine: nn.Module,
        params: PyTree,
        sampler_state: "sampler.SamplerState",  # noqa: F821
    ) -> Any | None:
        rule_states = []
        for i in range(len(self.probabilities)):
            # construct temporary sampler and rule state with correct sub-hilbert and
            # sampler-state objects.
            _state = sampler_state.replace(rule_state=sampler_state.rule_state[i])
            rule_states.append(self.rules[i].reset(sampler, machine, params, _state))
        return tuple(rule_states)

    def transition(self, sampler, machine, parameters, state, key, σ):
        N = len(self.probabilities)
        keys = jax.random.split(key, N + 1)

        σps = []
        log_prob_corrs = []
        for i in range(N):
            # construct temporary rule state with correct sampler-state objects
            _state = state.replace(rule_state=state.rule_state[i])

            σps_i, log_prob_corr_i = self.rules[i].transition(
                sampler, machine, parameters, _state, keys[i], σ
            )

            σps.append(σps_i)
            log_prob_corrs.append(log_prob_corr_i)

        indices = jax.random.choice(
            keys[-1],
            N,
            shape=(sampler.n_batches,),
            p=self.probabilities,
        )

        batch_select = jax.vmap(partial(jnp.take, axis=0))
        σp = batch_select(jnp.stack(σps, axis=1), indices)

        # if not all log_prob_corr are 0, convert the Nones to 0s
        if any(x is not None for x in log_prob_corrs):
            log_prob_corrs = jnp.stack(
                [
                    x if x is not None else jnp.zeros((sampler.n_batches,))
                    for x in log_prob_corrs
                ],
                axis=1,
            )
            log_prob_corr = batch_select(log_prob_corrs, indices)
        else:
            log_prob_corr = None

        return σp, log_prob_corr

    def __repr__(self):
        return f"MultipleRules(probabilities={self.probabilities}, rules={self.rules})"



# Copyright 2022-2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable
import warnings

import numpy as np
from fractions import Fraction

from netket.hilbert.homogeneous import HomogeneousHilbert
from netket.utils import StaticRange


from netket.hilbert.constraint import (
    DiscreteHilbertConstraint,
    ExtraConstraint,
    SumConstraint,
    SumOnPartitionConstraint,
)


class SpinOrbitalFermions(HomogeneousHilbert):
    r"""
    Hilbert space for 2nd quantization fermions with spin `s` distributed among
    `n_orbital` orbitals.

    The number of fermions can be fixed globally or fixed on a per spin projection.

    Note:
        This class is simply a convenient wrapper that creates a Fock or TensorHilbert
        of Fock spaces with occupation numbers 0 or 1.
        It is mainly useful to avoid needing to specify the n_max=1 each time, and adds
        convenient functions such as _get_index and _spin_index, which allow one to
        index the correct TensorHilbert corresponding to the right spin projection.
    """

    def __init__(
        self,
        n_orbitals: int,
        s: float | None = None,
        *,
        n_fermions: int | None = None,
        n_fermions_per_spin: tuple[int, ...] | None = None,
        constraint: DiscreteHilbertConstraint | None = None,
    ):
        r"""
        Constructs the hilbert space for spin-`s` fermions on `n_orbitals`.

        Samples of this hilbert space represent occupation numbers :math:`(0,1)` of the
        orbitals. The number of fermions may be fixed to `n_fermions`.
        If the spin is different from 0 or None, n_fermions can also be a list to fix
        the number of fermions per spin component.
        Using this class, one can generate a tensor product of fermionic hilbert spaces
        that distinguish particles with different spin.

        Args:
            n_orbitals: number of orbitals we store occupation numbers for. If the
                number of fermions per spin is conserved, the different spin
                configurations are not counted as orbitals and are handled differently.
            s: spin of the fermions.
            n_fermions: (optional) fixed number of fermions per spin (conserved). If specified,
                the total number of fermions is conserved. If the fermions have a spin, the
                number of fermions per spin subsector is not conserved.
            n_fermions_per_spin: (optional) list of the fixed number of fermions for every spin
                subsector. This automatically enforces a global population constraint. Only one
                between **n_fermions** and **n_fermions_per_spin** can be specified. The length
                of the iterable should be :math:`2S+1`.
            constraint: An extra constraint for the Hilbert space, defined according to the
                constraint API (see :class:`netket.hilbert.constraint.DiscreteHilbertConstraint`
                for more details).

        Returns:
            A SpinOrbitalFermions object
        """
        # TODO remove in 3.12 because it's deprecated experimental behaviour
        if isinstance(n_fermions, Iterable):
            warnings.warn(
                """
                          Declaring per-spin subsector constraint with `n_fermions` is deprecated.
                          Use `n_fermions_per_spin` instead.

                          As this class was experimental, this behaviour will break in NetKet 3.12 to be
                          released in early 2024.
                          """,
                DeprecationWarning,
                stacklevel=2,
            )
            n_fermions_per_spin = n_fermions
            n_fermions = None

        if not (isinstance(n_fermions, int) or n_fermions is None):
            raise TypeError(
                """
                The global `n_fermions` constraint must be an integer or be left unspecified.

                To declare a per-spin constraint, specify `n_fermions_per_spin` instead
                """
            )

        if s is None:
            spin_size = 1
        else:
            spin_size = round(2 * s + 1)

        total_size = n_orbitals * spin_size
        occupation_constraint = None

        if spin_size == 1:
            if n_fermions_per_spin is not None:
                raise ValueError(
                    "Cannot specify `n_fermions_per_spin` for spin-less " "fermions."
                )

            n_fermions_per_spin = (n_fermions,)

            if n_fermions is not None:
                occupation_constraint = SumConstraint(n_fermions)
        else:
            if n_fermions_per_spin is not None and n_fermions is not None:
                raise ValueError(
                    """
                                 Cannot specify the per-subsector constraint `n_fermions_per_spin`
                                 at the same time as the global constraint `n_fermions`.

                                 If you want to conserve the total number of fermions, but allow for
                                 changes in the number of fermions with a certain value of spin, you
                                 should specify `n_fermions`. If you want to fix the number of fermions
                                 with each spin component (and therefore the total number of fermions
                                 as well) you should specify `n_fermions_per_spin`.
                                 """
                )
            elif n_fermions_per_spin is not None:
                if not isinstance(n_fermions_per_spin, Iterable):
                    raise TypeError(
                        f"n_fermions_per_spin={n_fermions_per_spin} (whose type is {type(n_fermions_per_spin)}) "
                        "must be a list of integers or None describing the number of "
                        f"fermions in each of the {total_size} spin subsectors."
                    )
                if len(n_fermions_per_spin) != spin_size:
                    raise ValueError(
                        "List of number of fermions must equal number of spin components.\n"
                        f"For s={s}, which has {total_size} components, the same length is "
                        f"expected."
                    )
                n_fermions = sum(n_fermions_per_spin)

                # This is a special constraint on the subsectors of the hilbert space.
                # Equivalent to taking the kronecker product of the individual fock spaces.
                occupation_constraint = SumOnPartitionConstraint(
                    sum_values=tuple(n_fermions_per_spin),
                    sizes=tuple(n_orbitals for _ in n_fermions_per_spin),
                )
            else:
                # fixed fermion number but multiple spins subspaces
                # global or no constraint
                n_fermions_per_spin = tuple(None for _ in range(spin_size))
                if n_fermions is not None:
                    occupation_constraint = SumConstraint(n_fermions)

        if occupation_constraint is not None:
            # Wrap the extra user provided cosntraint around our
            # occupation constraint of the populations.
            if constraint is not None:
                constraint = ExtraConstraint(
                    base_constraint=occupation_constraint,
                    extra_constraint=constraint,
                )
            else:
                constraint = occupation_constraint

        """Internal representation of this Hilbert space (Fock or TensorHilbert)."""
        # local states are the occupation numbers (0, 1)
        local_states = StaticRange(0, 1, 2)

        # we use the constraints from the Fock spaces, and override `constrained`
        super().__init__(local_states, N=total_size, constraint=constraint)
        self._s = s
        self._n_fermions = n_fermions
        self._n_fermions_per_subsector: tuple[int | None, ...] = n_fermions_per_spin
        self._n_orbitals = n_orbitals

    @property
    def n_fermions(self) -> int | None:
        """The total number of fermions. None if unspecified."""
        return self._n_fermions

    @property
    def n_fermions_per_spin(self) -> tuple[int | None, ...]:
        """Tuple identifying the per-subsector population constraint.

        This tuple has length 1 for spinless fermions and length 2s+1 for spinful fermions.
        Every element is an integer or None, where None means no constraint on the number
        of fermions in that subsector.
        """
        return self._n_fermions_per_subsector

    @property
    def n_spin_subsectors(self) -> int:
        """Total number of spin subsectors. If spin is None, this is 1."""
        return len(self._n_fermions_per_subsector)

    @property
    def n_orbitals(self) -> int:
        """The number of orbitals (for each spin subsector if relevant)."""
        return self._n_orbitals

    @property
    def spin(self) -> float | None:
        """Returns the spin of the fermions"""
        return self._s

    @property
    def _attrs(self):
        return (self.n_orbitals, self.spin, self.constraint)

    @property
    def _n_spin_states(self) -> int:
        """Return the number of spin projections"""
        if self.spin is None:
            raise Exception(
                "cannot request number of spin states for spinless fermions"
            )
        return round(2 * self.spin + 1)

    def _spin_index(self, sz: int | None) -> int:
        """Return the index of the Fock block corresponding to the sz projection"""
        if self.spin is None:
            if sz is not None or not np.isclose(sz, 0):  # type: ignore[call-overload]
                raise Exception("cannot request spin index of spinless fermions")
            return 0
        else:
            if sz is None:
                raise TypeError("sz must be declared for spin-full hilbert spaces.")
            if sz != int(sz):
                raise TypeError(f"sz must be an integer, but got {type(sz)} ({sz})")
            if abs(sz) > 2 * self.spin:
                val = int(2 * self.spin)
                raise ValueError(
                    f"Valid spin values are in the interval [{-val},{val}]"
                )
            # if integer spin, valid values are even
            if int(self.spin) == self.spin:
                if sz % 2 != 0:
                    raise ValueError(
                        f"For spin S={self.spin}, valid spin values are odd"
                    )
            else:
                # valid values are odd
                if sz % 2 == 0:
                    raise ValueError(
                        f"For spin S={self.spin}, valid spin values are even"
                    )
            return (sz + int(2 * self.spin)) // 2

    def _get_index(self, orb: int, sz: int | None = None):
        """go from (site, spin_projection) indices to index in the hilbert space"""
        if orb >= self.n_orbitals:
            raise IndexError("requested orbital index outside of the hilbert space")
        spin_idx = self._spin_index(sz)
        return spin_idx * self.n_orbitals + orb

    def __repr__(self):
        _str = f"SpinOrbitalFermions(n_orbitals={self.n_orbitals}"
        if self.spin is not None:
            _str += f", s={Fraction(self.spin)}"
        if self.n_fermions is not None:
            _str += f", n_fermions={self.n_fermions}"
        if self.spin is not None and any(
            x is not None for x in self.n_fermions_per_spin
        ):
            _str += f", n_fermions_per_spin={self.n_fermions_per_spin}"
        _str += ")"
        return _str



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import ModuleType
import importlib


def import_optional_dependency(name: str, minimum_version="", descr="") -> ModuleType:
    """Try to import library `name`, and if it cannot be found, raise an
    informative error.
    """
    try:
        return importlib.import_module(name)
    except ModuleNotFoundError:
        if minimum_version != "":
            minimum_version = f">= {minimum_version}"
        raise ModuleNotFoundError(
            f"""

            Could not import `{name}`, which is necessary to use
            `{descr}`.

            To install it, run

                pip install {name} {minimum_version}

            """
        )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.utils import deprecated


@deprecated(reason="Will be removed. Please remove usages.")
def _info(obj, depth=None):
    if hasattr(obj, "info") and callable(obj.info):
        return obj.info(depth)
    else:
        return str(obj)


_deprecated_names = {"info": _info}


def __getattr__(name):
    from netket.utils import warn_deprecation

    if name in _deprecated_names:
        warn_deprecation(
            " \n"
            " \n"
            "          =======================================================================\n"
            "            `nk.driver.vmc_common is deprecated and the functionality removed.   \n"
            "          =======================================================================\n"
            " \n"
            "If you imported `nk.driver.vmc_common`, you must reimplement that functionality yourself.\n\n"
        )
        return _deprecated_names[name]

    raise AttributeError(f"module {__name__} has no attribute {name}")



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import flax.linen as nn
import jax.numpy as jnp

from functools import partial

import numpy as np

from netket.hilbert import SpinOrbitalFermions
from netket.utils.types import NNInitFunc, DType
from netket.nn.masked_linear import default_kernel_init
from netket import jax as nkjax


class Slater2nd(nn.Module):
    r"""
    A slater determinant ansatz for second-quantised spinless or spin-full
    fermions.

    When working with spin-full fermionic hilbert spaces (where the number of degrees of
    freedom is a multiple of the number of orbitals) :class:`~netket.models.Slater2nd`
    may behave in 3 different ways, depending on some flags.  Those modes differ by how
    the orbitals are represented.

    The more restrictions we impose, the lower the number of parameters,
    the higher the number of imposed symmetries, but the expressivity will be worse and it will
    be less likely to attain accurate ground-state energies.

    Those options are summarised in this table, while the details are discussed below.

    =================  =====================================================================  =============================================
    Hartree Fock type  Number of Parameters                                                   Options
    =================  =====================================================================  =============================================
    Generalized        :math:`n_{\mathrm{M}} \times n_{\mathrm{f}}`                           :code:`generalized = True`
    Unrestricted       :math:`n_{\mathrm{S}} \times n_{\mathrm{L}} \times n_{\textrm{f, s}}`  :code:`generalized=False, restricted=False`
    Restricted         :math:`n_{\mathrm{L}} \times n_{\textrm{f, s}}`                        :code:`generalized=False, restricted=True`
    =================  =====================================================================  =============================================


    Where we use :math:`n_{\textrm{M}}` to denote the number of fermionic modes, :math:`n_{\textrm{L}}`
    the number of spatial orbitals, and :math:`n_{\textrm{S}}` the number of spin states. The number
    of fermions is denoted :math:`n_{\mathrm{f}}`, or :math:`n_{\textrm{f}, \alpha}` for the number of
    fermions in a given spin sector :math:`\alpha`. We assume the same number of fermions in each spin
    sector for simplicity.


    Details about different hartree Fock types
    ==========================================

    Assume we introduce a set of orbitals :math:`\phi_\mu(r, s)` with orbital index :math:`\mu`.

    - **Generalized Hartree-Fock** (:code:`generalized = True`) is the most general case
      where we impose no restrictions. In particular, we do not restrict the orbitals
      to have definite spin or orbital quantum numbers.
      The total number of parameters is :math:`n_{\mathrm{M}} \times n_{\mathrm{f}}`. Hence,
      any fermion can occupy any of the fermionic modes.
    - **Hartree-Fock (Spin-Conserving)** (:code:`[generalized=False,] restricted=True/False`).
      Most physical Hamiltonians are spin conserving, and hence we can impose it also on the
      wave-function. In this case, we separate the orbital index :math:`\mu \to (l, \alpha)` into
      a spin and spatial orbital part: :math:`\phi_\mu(r, s)=\varphi_{l,\alpha}(r) \chi_{\alpha}(s)`.
      Here, :math:`l` and :math:`\alpha` indicate the orbital and spin quantum numbers associated
      with the orbital, and :math:`(r, s)` are the position vector and spin quantum number at which
      we aim to evaluate the orbital (i.e. properties of a given fermion). Furthermore,
      :math:`\varphi_{l,\alpha}(r)` is the spatial orbital at position :math:`r`, and and
      :math:`\chi_\alpha(s)` the spin part.

      - **Unrestricted Hartree Fock (UHF)** (:code:`[generalized=False,] restricted=False`),
        the orbitals can have a different spatial
        orbital :math:`\varphi` for different spin states. Since e.g. the up
        spin fermions cannot occupy the down spin orbitals and vice versa, the Slater matrix becomes block
        diagonal. This allows us to write the determinant as a product of determinants of the two spin sectors.
        The total number of parameters is :math:`n_{\mathrm{S}} \times n_{\mathrm{L}} \times n_{\textrm{f, s}}`. For
        more information, see
        `Wikipedia: Unrestricted Hartree-Fock <https://en.wikipedia.org/wiki/Unrestricted_Hartree%E2%80%93Fock>`_
      - **Restricted Hartree-Fock (RHF)** (:code:`[generalized=False, restricted=True]`), which assumes
        that different spin states have the same spatial orbitals
        in :math:`\phi_\mu(r, s)=\varphi_l(r) \chi_\alpha(s)`, and hence :math:`\varphi_l` only depends
        on the spatial orbital index :math:`l`. The number of
        parameters now reduces to :math:`n_{\mathrm{L}} \times n_{\textrm{f, s}}`.

    """

    hilbert: SpinOrbitalFermions
    """The Hilbert space upon which this ansatz is defined. Used to determine the number of orbitals
    and spin subspectors."""

    generalized: bool = False
    """Uses Generalized Hartree-Fock if True (defaults to `False`, corresponding to the
    standard spin-conserving Hartree-Fock).
    """

    restricted: bool = True
    """Flag to select the restricted- or unrestricted- Hartree Fock orbitals
    (Defaults to restricted).

    This flag is ignored if :code:`generalized=True`.

    - If restricted, only one set of orbitals are parametrised, and they are
      used for all spin subsectors. This only works if every spin subsector
      holds the same number of fermions.
    - If unrestricted, a different set of orbitals are parametrised and used
      for each spin subsector.
    """

    kernel_init: NNInitFunc = default_kernel_init
    """Initializer for the orbital parameters."""

    param_dtype: DType = float
    """Dtype of the orbital amplitudes."""

    def __post_init__(self):
        if not isinstance(self.hilbert, SpinOrbitalFermions):
            raise TypeError(
                "Slater2nd only supports 2nd quantised fermionic hilbert spaces."
            )
        if self.hilbert.n_fermions is None:
            raise TypeError(
                "Slater2nd only supports hilbert spaces with a "
                "fixed number of fermions."
            )
        if self.restricted:
            if not all(
                np.equal(
                    self.hilbert.n_fermions_per_spin,
                    self.hilbert.n_fermions_per_spin[0],
                )
            ):
                raise ValueError(
                    "Restricted Hartree Fock only makes sense for spaces with "
                    "same number of fermions on every subspace."
                )
        super().__post_init__()

    def setup(self):
        # Every determinant is a matrix of shape (n_orbitals, n_fermions_i) where
        # n_fermions_i is the number of fermions in the i-th spin sector.
        if self.generalized:
            M = self.param(
                "M",
                self.kernel_init,
                (self.hilbert.size, self.hilbert.n_fermions),
                self.param_dtype,
            )
            self.orbitals = M
        else:
            if self.restricted:
                M = self.param(
                    "M",
                    self.kernel_init,
                    (self.hilbert.n_orbitals, self.hilbert.n_fermions_per_spin[0]),
                    self.param_dtype,
                )

                self.orbitals = [M for _ in self.hilbert.n_fermions_per_spin]
            else:
                self.orbitals = [
                    self.param(
                        f"M_{i}",
                        self.kernel_init,
                        (self.hilbert.n_orbitals, nf_i),
                        self.param_dtype,
                    )
                    for i, nf_i in enumerate(self.hilbert.n_fermions_per_spin)
                ]

    def __call__(self, n):
        """
        Assumes inputs are strings of 0,1 that specify which orbitals are occupied.
        Spin sectors are assumed to follow the SpinOrbitalFermion's factorisation,
        meaning that the first `n_orbitals` entries correspond to sector -1, the
        second `n_orbitals` correspond to 0 ... etc.
        """
        if not n.shape[-1] == self.hilbert.size:
            raise ValueError(
                f"Dimension mismatch. Expected samples with {self.hilbert.size} "
                f"degrees of freedom, but got a sample of shape {n.shape} ({n.shape[-1]} dof)."
            )
        if not jnp.issubdtype(n, int):
            n = jnp.isclose(n, 1)

        @partial(jnp.vectorize, signature="(n)->()")
        def log_sd(n):
            # Find the positions of the occupied sites
            R = n.nonzero(size=self.hilbert.n_fermions)[0]
            log_det_sum = 0
            i_start = 0

            if self.generalized:
                # extract Nf x Nf submatrix
                A_i = self.orbitals[R, :]
                log_det_sum = nkjax.logdet_cmplx(A_i)
            else:
                for i, (n_fermions_i, M_i) in enumerate(
                    zip(self.hilbert.n_fermions_per_spin, self.orbitals)
                ):
                    if n_fermions_i == 0:
                        continue
                    # convert global orbital positions to spin-sector-local positions
                    R_i = (
                        R[i_start : i_start + n_fermions_i]
                        - i * self.hilbert.n_orbitals
                    )
                    # extract the corresponding Nf x Nf submatrix
                    A_i = M_i[R_i]

                    log_det_sum = log_det_sum + nkjax.logdet_cmplx(A_i)
                    i_start += n_fermions_i

            return log_det_sum

        return log_sd(n)


class MultiSlater2nd(nn.Module):
    r"""
    A slater determinant ansatz for second-quantised spinless or spin-full
    fermions with a sum of determinants.

    Refer to :class:`~netket.experimental.models.Slater2nd` for details about the different
    variants of Hartree Fock and the flags.
    """

    hilbert: SpinOrbitalFermions
    """The Hilbert space upon which this ansatz is defined. Used to determine the number of orbitals
    and spin subspectors."""

    n_determinants: int = 1
    """The number of determinants to be summed."""

    generalized: bool = False
    """Uses Generalized Hartree-Fock if True (defaults to `False`, corresponding to the
    standard spin-conserving Hartree-Fock).
    """

    restricted: bool = True
    """Flag to select the restricted- or unrestricted- Hartree Fock orbitals
    (Defaults to restricted).

    This flag is ignored if :code:`generalized=True`.

    - If restricted, only one set of orbitals are parametrised, and they are
      used for all spin subsectors. This only works if every spin subsector
      holds the same number of fermions.
    - If unrestricted, a different set of orbitals are parametrised and used
      for each spin subsector.
    """

    kernel_init: NNInitFunc = default_kernel_init
    """Initializer for the orbital parameters."""

    param_dtype: DType = float
    """Dtype of the orbital amplitudes."""

    @nn.compact
    def __call__(self, n):
        """
        Assumes inputs are strings of 0,1 that specify which orbitals are occupied.
        Spin sectors are assumed to follow the SpinOrbitalFermion's factorisation,
        meaning that the first `n_orbitals` entries correspond to sector -1, the
        second `n_orbitals` correspond to 0 ... etc.
        """
        if not self.n_determinants:
            raise ValueError(
                "Number of determinants must be an integer greater than 0."
            )
        # make extra axis with copies to run determinants in parallel
        n_bc = jnp.broadcast_to(n, (self.n_determinants, *n.shape))
        multi_log_det = nn.vmap(
            Slater2nd,
            in_axes=0,
            out_axes=0,  # vmap over copied axis
            variable_axes={"params": 0},
            split_rngs={"params": True},
        )(
            self.hilbert,
            restricted=self.restricted,
            generalized=self.generalized,
            kernel_init=self.kernel_init,
            param_dtype=self.param_dtype,
        )(
            n_bc
        )
        # sum the determinants
        log_det_sum = nkjax.logsumexp_cplx(multi_log_det, axis=0)
        return log_det_sum



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from netket.utils import struct
from netket.utils.types import Array, Scalar

max_states = np.iinfo(np.int32).max
"""int: Maximum number of states that can be indexed"""


def is_indexable(shape: tuple[Scalar] | Scalar) -> bool:
    """
    Returns whether a discrete Hilbert space of shape `shape` is
    indexable (i.e., its total number of states is below the maximum).
    """
    # cast to float to avoid
    # TypeError: loop of ufunc does not support argument 0 of type int which has no callable log method
    # when shape contains ints larger than the max int64
    shape = np.asarray(shape, dtype=np.float64)
    log_max = np.log(max_states)
    return np.sum(np.log(shape)) <= log_max


class HilbertIndex(struct.Pytree):
    @property
    def n_states(self) -> int:
        """Returns the size of the hilbert space."""
        return NotImplemented  # pragma: no cover

    @property
    def is_indexable(self) -> bool:
        """Whether the index can be indexed with an integer"""
        return NotImplemented  # pragma: no cover

    def states_to_numbers(self, states: Array) -> Array:
        """Given a Batch of N states of size M, returns an array
        of np.int32 integers used to numerate those states.
        """
        raise NotImplementedError  # pragma: no cover

    def numbers_to_states(self, numbers: Array) -> Array:
        """Given a list of N integers, returns a batch of N states of size M
        of corresponding to those states. This function is the inverse
        of `states_to_numbers`.
        """
        raise NotImplementedError  # pragma: no cover

    def all_states(self) -> Array:
        """Should behave as `self.numbers_to_states(range(self.n_states))`
        but might be optimised for iterating across the full hilbert
        space.
        """
        raise NotImplementedError  # pragma: no cover



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

from flax.linen import Module

import jax.numpy as jnp


from .partial import HashablePartial
from .types import ModuleOrApplyFun, PyTree, Array

from . import struct


def get_afun_if_module(mod_or_fun) -> Callable:
    """Returns the apply function if it's a module. Does nothing otherwise."""
    if hasattr(mod_or_fun, "apply"):
        return mod_or_fun.apply
    else:
        return mod_or_fun


@struct.dataclass
class WrappedApplyFun:
    """Wraps a callable to be a module-like object with the method `apply`."""

    apply: Callable[[PyTree, Array], Array]
    """The wrapped callable."""

    def __repr__(self):
        return f"{type(self).__name__}(apply={self.apply}, hash={hash(self)})"


def wrap_afun(mod_or_fun: ModuleOrApplyFun) -> Module:
    """Wraps a callable to be a module-like object with the method `apply`.
    Does nothing if it already has an apply method.
    """
    if hasattr(mod_or_fun, "apply"):
        return mod_or_fun  # type: ignore
    else:
        return WrappedApplyFun(mod_or_fun)  # type: ignore[return-type]


def wrap_to_support_scalar(fun):
    """
    Wraps the flax-compatible apply function, assuming that the state input is the
    second argument, so that it always calls the wrapped function with a tensor with at
    least 2 dimensions.

    If the input was 1-dimensional, returns a scalar instead of a vector with 1 element.

    DEVNOTE: This function is used because some parts of NetKet make use of the fact
    that when we call the logψ with a single bitstring, we get a scalar out. This is
    useful when calling `jax.grad(logψ)`.
    This also makes sure that users only need to write networks that work on batches,
    and not necessarily on scalars.

    Args:
        fun: A flax-compatible function.

    Returns:
        A wrapped function, returned as an `HashablePartial` in order not to retrigger
        compilation.
    """

    def maybe_scalar_fun(apply_fun, pars, x, *args, **kwargs):
        xb = jnp.atleast_2d(x)
        res = apply_fun(pars, xb, *args, **kwargs)
        # support models with mutable state
        if isinstance(res, tuple):
            res_val = res[0]
            res_val = res_val.reshape(()) if x.ndim == 1 else res_val
            res = (res_val, res[1])
        else:
            res = res.reshape(()) if x.ndim == 1 else res
        return res

    return HashablePartial(maybe_scalar_fun, fun)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from collections.abc import Callable

import jax
from jax import numpy as jnp
from flax.core.scope import CollectionFilter, DenyList  # noqa: F401

from netket import jax as nkjax
from netket import config
from netket.stats import Stats
from netket.utils import mpi, dispatch
from netket.utils.types import PyTree

from netket.operator import (
    AbstractOperator,
    Squared,
)

from netket.vqs import expect_and_grad, expect_and_forces

from ..common import force_to_grad, get_local_kernel_arguments, get_local_kernel

from .state import MCState


# General implementation checking hermitianity
@expect_and_grad.dispatch
def expect_and_grad_default_formula(
    vstate: MCState,
    Ô: AbstractOperator,
    chunk_size: int | None,
    *args,
    mutable: CollectionFilter = False,
    use_covariance: bool | None = None,
) -> tuple[Stats, PyTree]:
    if use_covariance is None:
        use_covariance = Ô.is_hermitian

    if use_covariance:
        # Implementation of expect_and_grad for `use_covariance == True` (due to the Literal[True]
        # type in the signature).` This case is equivalent to the composition of the
        # `expect_and_forces` and `force_to_grad` functions.
        # return expect_and_grad_from_covariance(vstate, Ô, *args, mutable=mutable)
        Ō, Ō_grad = expect_and_forces(vstate, Ô, chunk_size, *args, mutable=mutable)
        Ō_grad = force_to_grad(Ō_grad, vstate.parameters)
        return Ō, Ō_grad
    else:
        return expect_and_grad_nonhermitian(
            vstate, Ô, chunk_size, *args, mutable=mutable
        )


# Squared is a special operator...
@expect_and_grad.dispatch
def expect_and_grad_squared_op(
    vstate: MCState,
    Ô: Squared,
    chunk_size: int | None,
    *args,
    mutable: CollectionFilter = False,
    use_covariance: bool | None = None,
) -> tuple[Stats, PyTree]:
    if use_covariance is not None:
        raise ValueError(
            "Cannot specify `use_covariance` with Squared[...] operator.\n"
            "This operator must use the same formula as non-hermitian operators to work."
        )
    return expect_and_grad_nonhermitian(vstate, Ô, chunk_size, *args, mutable=mutable)


@dispatch.dispatch
def expect_and_grad_nonhermitian(
    vstate: MCState,
    Ô,
    chunk_size: None,
    *,
    mutable: CollectionFilter = False,
):
    if not isinstance(Ô, Squared) and not config.netket_experimental:
        raise RuntimeError(
            """
            Computing the gradient of non hermitian operator is an
            experimental feature under development and is known not to
            return wrong values sometimes.

            If you want to debug it, set the environment variable
            NETKET_EXPERIMENTAL=1
            """
        )

    σ, args = get_local_kernel_arguments(vstate, Ô)

    local_estimator_fun = get_local_kernel(vstate, Ô)

    Ō, Ō_grad, new_model_state = _grad_expect_nonherm_kernel(
        local_estimator_fun,
        vstate._apply_fun,
        mutable,
        vstate.sampler.machine_pow,
        vstate.parameters,
        vstate.model_state,
        σ,
        args,
    )

    if mutable is not False:
        vstate.model_state = new_model_state

    return Ō, Ō_grad


@partial(jax.jit, static_argnums=(0, 1, 2))
def _grad_expect_nonherm_kernel(
    local_value_kernel: Callable,
    model_apply_fun: Callable,
    mutable: CollectionFilter,
    machine_pow: float,
    parameters: PyTree,
    model_state: PyTree,
    σ: jnp.ndarray,
    local_value_args: PyTree,
) -> tuple[PyTree, PyTree, Stats]:
    n_chains = σ.shape[0]
    if σ.ndim >= 3:
        σ = jax.lax.collapse(σ, 0, 2)

    is_mutable = mutable is not False
    logpsi = lambda w, σ: model_apply_fun(
        {"params": w, **model_state}, σ, mutable=mutable
    )
    log_pdf = (
        lambda w, σ: machine_pow * model_apply_fun({"params": w, **model_state}, σ).real
    )

    def expect_closure_pars(pars):
        return nkjax.expect(
            log_pdf,
            partial(local_value_kernel, logpsi),
            pars,
            σ,
            local_value_args,
            n_chains=n_chains,
        )

    Ō, Ō_pb, Ō_stats = nkjax.vjp(
        expect_closure_pars, parameters, has_aux=True, conjugate=True
    )
    Ō_pars_grad = Ō_pb(jnp.ones_like(Ō))[0]
    Ō_pars_grad, _ = mpi.mpi_sum_jax(Ō_pars_grad)

    if is_mutable:
        raise NotImplementedError(
            "gradient of non-hermitian operators over mutable models "
            "is not yet implemented."
        )
    new_model_state = None

    return (
        Ō_stats,
        Ō_pars_grad,
        new_model_state,
    )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import jax
import jax.numpy as jnp

from netket.hilbert import AbstractHilbert
from netket.utils.dispatch import dispatch


def check_hilbert(A: AbstractHilbert, B: AbstractHilbert):
    if not A == B:
        raise NotImplementedError(  # pragma: no cover
            f"Non matching hilbert spaces {A} and {B}"
        )


@dispatch.abstract
def get_local_kernel_arguments(vstate: Any, Ô: Any):
    """
    Returns the samples of vstate used to compute the expectation value
    of the operator O, and the connected elements and matrix elements.

    Args:
        vstate: the variational state
        Ô: the operator

    Returns:
        A Tuple with 2 elements (sigma, args), where the first elements
        should be the samples over which the classical expectation value
        should be computed, while the latter is anything that can be fed
        as input to the local_kernel.
    """


@dispatch.abstract
def get_local_kernel(vstate: Any, Ô: Any):
    """
    Returns the function computing the local estimator for the given variational
    state and operator.

    Args:
        vstate: the variational state
        Ô: the operator

    Returns:
        A callable accepting the output of `get_configs(vstate, O)`.
    """


@jax.jit
def force_to_grad(Ō_grad, parameters):
    """
    Converts the forces vector F_k = cov(O_k, E_loc) to the observable gradient.
    In case of a complex target (which we assume to correspond to a holomorphic
    parametrization), this is the identity. For real-valued parameters, the gradient
    is 2 Re[F].
    """
    Ō_grad = jax.tree_util.tree_map(
        lambda x, target: (x if jnp.iscomplexobj(target) else x.real).astype(
            target.dtype
        ),
        Ō_grad,
        parameters,
    )
    Ō_grad = jax.tree_util.tree_map(lambda x: 2 * x, Ō_grad)
    return Ō_grad



import numpy as np
import sparse

from typing import Sequence
from netket.graph import AbstractGraph
from netket.utils.types import DType

import jax.numpy as jnp

from netket.utils import struct
from netket.hilbert import SpinOrbitalFermions

from . import ParticleNumberAndSpinConservingFermioperator2nd


@struct.dataclass
class FermiHubbardJax(ParticleNumberAndSpinConservingFermioperator2nd):
    r"""
    Fermi-Hubbard Hamiltonian based on the generic ParticleNumberAndSpinConservingFermioperator2nd

    .. math::
        \hat H = -t \sum_{<ij>,\sigma} (\hat c_{i\sigma}^\dagger \hat c_{j\sigma} + h.c.) + U \sum_{i}(\hat n_{i\uparrow} \hat n_{i\downarrow})

    This implementation is more efficient than a FermionOperator2nd created using create,destroy,number from nk.operator.fermion
    as shown in the examples, as it considers only matrix elements with the correct number of particles.
    """

    def __pre_init__(
        self,
        hilbert: SpinOrbitalFermions,
        graph: AbstractGraph,
        t: float | Sequence[float] = 1.0,
        U: float | Sequence[float] = 1.0,
        dtype: DType | None = None,
    ):
        r"""
        Constructs a new FermiHubbardJax operator given a hilbert space, a graph
        specifying the connectivity and the interaction strength.

        Args:
           hilbert: Hilbert space the operator acts on.
           graph: Graph
           t: The strength of the hopping term.
           U: The strength of the on-site coulomb interaction term.
           dtype: The dtype of the matrix elements.

        Examples:
           Constructs a FermiHubbardJax operator for a 2D system at half filling.

           >>> import netket as nk
           >>> import netket.experimental
           >>> g = nk.graph.Hypercube(length=4, n_dim=2, pbc=True)
           >>> hi = nk.hilbert.SpinOrbitalFermions(n_orbitals=g.n_nodes, s=1/2, n_fermions_per_spin=(4,4))
           >>> op = netket.experimental.operator.FermiHubbardJax(hi, t=1.0, U=1.0, graph=g)
        """

        if isinstance(t, Sequence):
            assert len(t) == graph.n_edges
        if isinstance(U, Sequence):
            assert len(U) == graph.n_nodes
        assert hilbert.n_spin_subsectors == 2
        assert hilbert.size == 2 * graph.n_nodes

        t = jnp.asarray(t, dtype=dtype)
        U = jnp.asarray(U, dtype=dtype)

        ij = np.array(graph.edges()).T
        t = np.broadcast_to(t, graph.n_edges)
        t_mat = sparse.COO(ij, t, shape=(graph.n_nodes,) * 2)

        U = np.broadcast_to(U, graph.n_nodes)
        iiii = np.array(
            [
                np.array(graph.nodes()),
            ]
            * 4
        )
        U_mat = sparse.COO(iiii, U, shape=(graph.n_nodes,) * 4)

        operators_sector = {}
        operators_sector[2, (0, 1)] = -(t_mat + t_mat.T)
        operators_sector[4, ((1, 0),)] = U_mat

        op = ParticleNumberAndSpinConservingFermioperator2nd._from_sparse_arrays_normal_order(
            hilbert, operators_sector
        )

        return (op._hilbert, op._operator_data), {}

    @property
    def is_hermitian(self):
        return True



# Copyright 2024 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from typing import TypeVar, ParamSpec

import time
import inspect
import functools
import contextlib

import jax

from rich.tree import Tree
from rich.panel import Panel

from netket.utils import config, struct, display

T = TypeVar("T")
P = ParamSpec("P")


CURRENT_TIMER_STACK = []


@display.rich_repr
class Timer(struct.Pytree, mutable=True):
    """
    Measures how much time was spent in a timer, with possible
    sub-timers.

    Can be used as a context manager to time a scope, and will show
    up timers used inside of the scope if present.

    If you directly construct a timer, you cannot nest it inside another
    timer. If you are building a library function you should instead use
    :func:`netket.utils.timing.timed_scope`.

    Example:

        Time a scope

        >>> import netket as nk
        >>> import time
        >>>
        >>> with nk.utils.timing.Timer() as t:
        ...    time.sleep(1)  # This line and the ones below are indented
        ...    with nk.utils.timing.timed_scope("subfunction 1"):
        ...       time.sleep(0.5)
        ...    with nk.utils.timing.timed_scope("subfunction 2"):
        ...       time.sleep(0.25)
        >>>
        >>> t  # doctest: +SKIP
        ╭──────────────────────── Timing Information ─────────────────────────╮
        │ Total: 1.763                                                        │
        │ ├── (28.7%) | subfunction 1 : 0.505 s                               │
        │ └── (14.3%) | subfunction 2 : 0.252 s                               │
        ╰─────────────────────────────────────────────────────────────────────╯

    """

    total: float
    sub_timers: dict
    _start_time: float = 0.0

    def __init__(self):
        """
        Constructs a new timer object.

        Does not accept any argument.
        """
        self.total = 0.0
        self.sub_timers = {}

    def __rich__(self, indent=0, total_above=None):
        # The initial part of the string representation for the current object

        tree = Tree(f"Total: {self.total:.3f}")
        self._rich_walk_tree_(tree)
        return Panel(tree, title="Timing Information")

    def _rich_walk_tree_(self, tree):
        attributed = 0.0
        for key, sub_timer in self.sub_timers.items():
            if self.total > 0 and sub_timer.total / self.total > 0.01:
                percentage = 100 * (sub_timer.total / self.total)
                attributed += sub_timer.total

                sub_tree = tree.add(
                    f"({percentage:.1f}%) | {key} : {sub_timer.total:.3f} s"
                )
                sub_timer._rich_walk_tree_(sub_tree)

        # This prints an 'other' line
        # if attributed > 0 and attributed / self.total < 0.98:
        #    rest = self.total - attributed
        #    percentage = 100 * (rest / self.total)
        #    tree.add(f"({percentage:.1f}%) | other : {rest:.3f} s")

    def __iadd__(self, o):
        if not isinstance(o, Timer):
            return NotImplemented
        self.total += o.total
        for k, v in o.sub_timers.items():
            if k in self.sub_timers:
                self.sub_timers[k] += v
            else:
                self.sub_timers[k] = v
        return self

    def get_subtimer(self, name: str):
        if name not in self.sub_timers:
            self.sub_timers[name] = Timer()
        return self.sub_timers[name]

    def block_until_ready(self, args):
        return jax.block_until_ready(args)

    def __enter__(self):
        global CURRENT_TIMER_STACK
        CURRENT_TIMER_STACK.append(self)
        self._start_time = time.perf_counter()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        t_end = time.perf_counter()
        current_timer = t_end - self._start_time
        self.total += current_timer
        _ = CURRENT_TIMER_STACK.pop()
        assert _ is self


class NullTimer(struct.Pytree):
    """
    A timer-compatible class that does nothing.

    Used to return
    """

    def get_subtimer(self, name: str):
        return self

    def block_until_ready(self, args):
        return args

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


_NULLTIMER = NullTimer()


@contextlib.contextmanager
def timed_scope(name: str = None, force: bool = False):
    """
    Context manager used to mark a scope to be timed individually
    by NetKet timers.

    If name is not specified, the file name and line number
    is used.

    If `force` is not specified, the timer only runs if a top-level
    timer is in use as well. If `force` is specified, the timer
    and nested timers will always run.

    .. note::

        If you are using JAX functions in your code, to get reliable timing
        results you should block the output of the function with `block_until_ready`.
        As this can sensibly slow down your code, this is only done when `force=True`
        or if a top-level timer is in use.

        To block the output on those conditions, simply use `timer.block_until_ready(output)`
        on the output of the function you want to time.

    Example:

        Time a section of code

        >>> import netket as nk
        >>> import time; import jax
        >>>
        >>> with nk.utils.timing.timed_scope(force=True) as timer:
        ...    time.sleep(1)  # This line and the ones below are indented
        ...    with nk.utils.timing.timed_scope("subfunction 1"):
        ...       time.sleep(0.5)
        ...    with nk.utils.timing.timed_scope("subfunction 2"):
        ...       time.sleep(0.25)
        ...    a = jax.random.normal(jax.random.key(1), (100,100))
        ...    # Must block jax functions otherwise the timing is off
        ...    timer.block_until_ready(timer)
        ...    # doctest: +SKIP
        >>>
        >>> timer  # doctest: +SKIP
        ╭──────────────────────── Timing Information ─────────────────────────╮
        │ Total: 1.763                                                        │
        │ ├── (28.7%) | subfunction 1 : 0.505 s                               │
        │ └── (14.3%) | subfunction 2 : 0.252 s                               │
        ╰─────────────────────────────────────────────────────────────────────╯



    Args:
        name: Name to use for the timing of this line.
        force: whether to always time, even if no top level timer
            is in use
    """
    __tracebackhide__ = True

    global CURRENT_TIMER_STACK
    if force or len(CURRENT_TIMER_STACK) > 0:
        if name is None:
            # If no name specified look 2 frames up (1 frame is
            # @contextmanager, 2 frames is caller) to get filename and
            # line number.
            caller_frame = inspect.stack()[2]
            frame = caller_frame[0]
            info = inspect.getframeinfo(frame)
            name = f"{info.filename}:{info.lineno}"

        if len(CURRENT_TIMER_STACK) == 0:
            timer = Timer()
        else:
            timer = CURRENT_TIMER_STACK[-1].get_subtimer(name)
        with timer:
            yield timer
    else:  # disabled
        yield _NULLTIMER


def timed(
    fun: Callable[P, T] = None, name: str | None = None, block_until_ready: bool = True
) -> Callable[P, T]:
    """
    Marks the decorated function to be timed individually in
    NetKet timing scopes.

    If name is not specified, the qualified name of the
    function is used.

    The profiling is disabled if no global timer is active.

    Args:
        fun: Function to be decorated
        name: Name to use for the timing of this line.
        block_until_ready: Calls `jax.block_until_ready` on the output if timing (default True),
            which slows down the code but gives accurate timing results of JAX functions.
    """
    if fun is None:
        return functools.partial(timed, name=name)

    if name is None:
        if hasattr(fun, "__qualname__"):
            name = fun.__qualname__
        else:
            name = fun.__name__

    @functools.wraps(fun)
    def timed_function(*args, **kwargs):
        __tracebackhide__ = True
        with timed_scope(name) as ts:
            result = fun(*args, **kwargs)
            if block_until_ready:
                return ts.block_until_ready(result)
            else:
                return result

    # Without this the link to code in sphinx is broken
    if config.netket_sphinx_build:
        return fun

    return timed_function


def clear_timers():
    """
    Resets all timers.
    """
    global CURRENT_TIMER_STACK
    CURRENT_TIMER_STACK.clear()



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from textwrap import dedent

from netket.utils import timing
from netket.utils.types import PyTree, Optimizer
from netket.operator import AbstractOperator
from netket.stats import Stats
from netket.optimizer import (
    identity_preconditioner,
    PreconditionerT,
)
from netket.vqs import VariationalState
from netket.jax import tree_cast

from .abstract_variational_driver import AbstractVariationalDriver


class VMC(AbstractVariationalDriver):
    """
    Energy minimization using Variational Monte Carlo (VMC).
    """

    def __init__(
        self,
        hamiltonian: AbstractOperator,
        optimizer: Optimizer,
        *,
        variational_state: VariationalState,
        preconditioner: PreconditionerT = identity_preconditioner,
    ):
        """
        Initializes the driver class.

        Args:
            hamiltonian: The Hamiltonian of the system.
            optimizer: Determines how optimization steps are performed given the
                bare energy gradient.
            variational_state: The variational state for which the hamiltonian must
                be minimised.
            preconditioner: Determines which preconditioner to use for the loss gradient.
                This must be a tuple of `(object, solver)` as documented in the section
                `preconditioners` in the documentation. The standard preconditioner
                included with NetKet is Stochastic Reconfiguration. By default, no
                preconditioner is used and the bare gradient is passed to the optimizer.
        """
        if variational_state.hilbert != hamiltonian.hilbert:
            raise TypeError(
                dedent(
                    f"""the variational_state has hilbert space {variational_state.hilbert}
                    (this is normally defined by the hilbert space in the sampler), but
                    the hamiltonian has hilbert space {hamiltonian.hilbert}.
                    The two should match.
                    """
                )
            )

        super().__init__(variational_state, optimizer, minimized_quantity_name="Energy")

        self._ham = hamiltonian.collect()  # type: AbstractOperator

        self.preconditioner = preconditioner

        self._dp: PyTree = None
        self._S = None

    @property
    def preconditioner(self):
        """
        The preconditioner used to modify the gradient.

        This is a function with the following signature

        .. code-block:: python

            precondtioner(vstate: VariationalState,
                          grad: PyTree,
                          step: Optional[Scalar] = None)

        Where the first argument is a variational state, the second argument
        is the PyTree of the gradient to precondition and the last optional
        argument is the step, used to change some parameters along the
        optimisation.

        Often, this is taken to be :func:`~netket.optimizer.SR`. If it is
        set to `None`, then the identity is used.
        """
        return self._preconditioner

    @preconditioner.setter
    def preconditioner(self, val: PreconditionerT | None):
        if val is None:
            val = identity_preconditioner

        self._preconditioner = val

    @timing.timed
    def _forward_and_backward(self):
        """
        Performs a number of VMC optimization steps.

        Args:
            n_steps (int): Number of steps to perform.
        """

        self.state.reset()

        # Compute the local energy estimator and average Energy
        self._loss_stats, self._loss_grad = self.state.expect_and_grad(self._ham)

        # if it's the identity it does
        # self._dp = self._loss_grad
        self._dp = self.preconditioner(self.state, self._loss_grad, self.step_count)

        # If parameters are real, then take only real part of the gradient (if it's complex)
        self._dp = tree_cast(self._dp, self.state.parameters)

        return self._dp

    @property
    def energy(self) -> Stats:
        """
        Return MCMC statistics for the expectation value of observables in the
        current state of the driver.
        """
        return self._loss_stats

    def __repr__(self):
        return (
            "Vmc("
            + f"\n  step_count = {self.step_count},"
            + f"\n  state = {self.state})"
        )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from functools import partial

import jax
from jax import numpy as jnp

from netket.stats import Stats, statistics as mpi_statistics
from netket.utils.types import PyTree
from netket.utils.dispatch import dispatch

from netket.operator import (
    AbstractOperator,
    DiscreteOperator,
    Squared,
    ContinuousOperator,
    DiscreteJaxOperator,
)

from netket.vqs.mc import (
    kernels,
    check_hilbert,
    get_local_kernel_arguments,
    get_local_kernel,
)

from .state import MCState


@dispatch
def get_local_kernel_arguments(vstate: MCState, Ô: Squared):  # noqa: F811
    check_hilbert(vstate.hilbert, Ô.hilbert)

    σ = vstate.samples
    σp, mels = Ô.parent.get_conn_padded(σ)
    return σ, (σp, mels)


@dispatch
def get_local_kernel(vstate: MCState, Ô: Squared):  # noqa: F811
    return kernels.local_value_squared_kernel


@dispatch
def get_local_kernel_arguments(vstate: MCState, Ô: DiscreteOperator):  # noqa: F811
    check_hilbert(vstate.hilbert, Ô.hilbert)

    σ = vstate.samples
    σp, mels = Ô.get_conn_padded(σ)
    return σ, (σp, mels)


@dispatch
def get_local_kernel(vstate: MCState, Ô: DiscreteOperator):  # noqa: F811
    return kernels.local_value_kernel


@dispatch
def get_local_kernel_arguments(vstate: MCState, Ô: DiscreteJaxOperator):  # noqa: F811
    check_hilbert(vstate.hilbert, Ô.hilbert)

    σ = vstate.samples
    return σ, Ô


@dispatch
def get_local_kernel(vstate: MCState, Ô: DiscreteJaxOperator):  # noqa: F811
    return kernels.local_value_kernel_jax


@dispatch
def get_local_kernel_arguments(vstate: MCState, Ô: ContinuousOperator):  # noqa: F811
    check_hilbert(vstate.hilbert, Ô.hilbert)

    σ = vstate.samples
    args = Ô._pack_arguments()
    return σ, args


@dispatch
def get_local_kernel(vstate: MCState, Ô: ContinuousOperator):  # noqa: F811
    # TODO: this should be moved other to dispatch in order to support MCMixedState
    return Ô._expect_kernel


# Standard implementation of expect for an MCState (pure) and a generic operator
# The dispatch rule is not strictly needed, as everything currently implemented
# in NetKet only defines a custom get_local_kernel_arguments and get_local_kernel
# but if somebody wants to override behaviour for an existing operator or define
# a completely arbitrary novel type of operator, this makes it much easier.
@dispatch
def expect(
    vstate: MCState, Ô: AbstractOperator, chunk_size: None
) -> Stats:  # noqa: F811
    σ, args = get_local_kernel_arguments(vstate, Ô)
    local_estimator_fun = get_local_kernel(vstate, Ô)

    return _expect(
        local_estimator_fun,
        vstate._apply_fun,
        vstate.sampler.machine_pow,
        vstate.parameters,
        vstate.model_state,
        σ,
        args,
    )


@partial(jax.jit, static_argnums=(0, 1))
def _expect(
    local_value_kernel: Callable,
    model_apply_fun: Callable,
    machine_pow: int,
    parameters: PyTree,
    model_state: PyTree,
    σ: jnp.ndarray,
    local_value_args: PyTree,
) -> Stats:
    n_chains = σ.shape[0]
    if σ.ndim >= 3:
        σ = jax.lax.collapse(σ, 0, 2)

    def logpsi(w, σ):
        return model_apply_fun({"params": w, **model_state}, σ)

    def log_pdf(w, σ):
        return machine_pow * model_apply_fun({"params": w, **model_state}, σ).real

    # TODO: Broken until google/jax#11916 is resolved.
    # should uncomment and remove code below once this is fixed
    # _, Ō_stats = nkjax.expect(
    #    log_pdf,
    #    partial(local_value_kernel, logpsi),
    #    parameters,
    #    σ,
    #    local_value_args,
    #    n_chains=n_chains,
    # )

    L_σ = local_value_kernel(logpsi, parameters, σ, local_value_args)
    Ō_stats = mpi_statistics(L_σ.reshape((n_chains, -1)))

    return Ō_stats



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from collections.abc import Callable
from collections.abc import Iterator
from typing import overload, Literal

import jax
from jax import numpy as jnp
from flax import linen as nn


from netket import jax as nkjax
from netket.jax import sharding
from netket import config
from netket.hilbert import AbstractHilbert, HomogeneousHilbert
from netket.utils import get_afun_if_module, struct, wrap_afun
from netket.utils.types import PyTree, DType, SeedT, ModuleOrApplyFun
from netket.jax import HashablePartial


class SamplerState(struct.Pytree):
    """
    Base class holding the state of a sampler.
    """


class Sampler(struct.Pytree):
    """
    Abstract base class for all samplers.

    It contains the fields that all of them should possess, defining the common
    API.
    Note that fields marked with `pytree_node=False` are treated as static arguments
    when jitting.

    Subclasses should be NetKet dataclasses and they should define the `_init_state`,
    `_reset` and `_sample_chain` methods which only accept positional arguments.
    See the respective method's definition for its signature.

    Notice that those methods are different from the API-entry point without the leading
    underscore in order to allow us to share some pre-processing code between samplers
    and simplify the definition of a new sampler.
    """

    hilbert: AbstractHilbert = struct.field(pytree_node=False)
    """The Hilbert space to sample."""

    machine_pow: float = struct.field(default=2.0)
    """The power to which the machine should be exponentiated to generate the pdf."""

    dtype: DType = struct.field(pytree_node=False, default=None)
    """The dtype of the states sampled."""

    def __init__(
        self,
        hilbert: AbstractHilbert,
        *,
        machine_pow: float = 2,
        dtype: DType = float,
    ):
        """
        Construct a Monte Carlo sampler.

        Args:
            hilbert: The Hilbert space to sample.
            machine_pow: The power to which the machine should be exponentiated to generate the pdf (default = 2).
            dtype: The dtype of the states sampled (default = np.float64).
        """
        # Raise errors if hilbert is not an Hilbert
        if not isinstance(hilbert, AbstractHilbert):
            raise TypeError(
                "\n\nThe argument `hilbert` of a Sampler must be a subtype "
                "of netket.hilbert.AbstractHilbert, but you passed in an object "
                f"of type {type(hilbert)}, which is not an AbstractHilbert.\n\n"
                "TO FIX THIS ERROR,\ndouble check the arguments passed to the "
                "sampler when constructing it, and verify that they have the "
                "correct types.\n\n"
                "For more information, check the correct arguments in the API "
                "reference at https://netket.readthedocs.io/en/latest/api/sampler.html"
                "\n"
            )

        if dtype is None:
            # If dtype is not specified, if it is an Homogeneous Hilbert we can just get it off
            # from the local states (StaticRange) otherwise we need to evaluate the shape of the
            # result of `jax.random_state` to automatically determine it.
            if isinstance(hilbert, HomogeneousHilbert):
                dtype = hilbert._local_states.dtype
            else:
                dtype = jax.eval_shape(
                    hilbert.random_state, jax.eval_shape(jax.random.key, 1)
                ).dtype

        machine_pow = jnp.array(machine_pow)

        if jnp.issubdtype(machine_pow, jnp.complexfloating):
            raise ValueError(f"machine_pow ({machine_pow}) must be real")

        # Below we want to check that machine_pow is positive, but in a way
        # that works also for samplers constructed inside a jit-context.
        # To make this work, we could use equinox.error_if, but we assume users
        # are smart enough and we only check if we are outside of jit
        if not isinstance(machine_pow, jax.core.Tracer):
            if machine_pow < 0:
                raise ValueError(f"machine_pow ({machine_pow}) must be positive")
        # else: equinox.error_if(machine_pow, machine_pow<0, ...)

        self.hilbert = hilbert
        self.machine_pow = machine_pow
        self.dtype = dtype

    @property
    def n_chains_per_rank(self) -> int:
        """
        The total number of independent chains per MPI rank (or jax device
        if you set `NETKET_EXPERIMENTAL_SHARDING=1`).

        If you are not distributing the calculation among different MPI ranks
        or jax devices, this is equal to :attr:`~Sampler.n_chains`.

        In general this is equal to

        .. code:: python

            from netket.jax import sharding
            sampler.n_chains // sharding.device_count()

        """
        n_devices = sharding.device_count()
        res, remainder = divmod(self.n_chains, n_devices)

        if remainder != 0:
            raise RuntimeError(
                "The number of chains is not a multiple of the number of mpi ranks"
            )
        return res

    @property
    def n_chains(self) -> int:
        """
        The total number of independent chains.

        This is at least equal to the total number of MPI ranks/jax devices that
        are used to distribute the calculation.
        """
        # This is the default number of chains, intended for generic non-mcmc
        # samplers which don't have a concept of chains.
        # We assume there is 1 dummy chain per mpi rank / jax device.
        # Currently this is used by the exact samplers (ExactSampler, ARDirectSampler).
        return sharding.device_count()

    @property
    def n_batches(self) -> int:
        r"""
        The batch size of the configuration $\sigma$ used by this sampler on this
        jax process.

        This is used to determine the shape of the batches generated in a single process.
        This is needed because when using MPI, every process must create a batch of chains
        of :attr:`~Sampler.n_chains_per_rank`, while when using the experimental sharding
        mode we must declare the full shape on every jax process, therefore this returns
        :attr:`~Sampler.n_chains`.

        Usage of this flag is required to support both MPI and sharding.

        Samplers may override this to have a larger batch size, for example to
        propagate multiple replicas (in the case of parallel tempering).
        """
        if config.netket_experimental_sharding:
            return self.n_chains
        else:
            return self.n_chains_per_rank

    @property
    def is_exact(self) -> bool:
        """
        Returns `True` if the sampler is exact.

        The sampler is exact if all the samples are exactly distributed according to the
        chosen power of the variational state, and there is no correlation among them.
        """
        return False

    def log_pdf(self, model: Callable | nn.Module) -> Callable:
        """
        Returns a closure with the log-pdf function encoded by this sampler.

        Args:
            model: A Flax module or callable with the forward pass of the log-pdf.
                If it is a callable, it should have the signature :code:`f(parameters, σ) -> jax.Array`.

        Returns:
            The log-probability density function.

        Note:
            The result is returned as a `HashablePartial` so that the closure
            does not trigger recompilation.
        """
        apply_fun = get_afun_if_module(model)
        log_pdf = HashablePartial(
            lambda apply_fun, pars, σ: self.machine_pow * apply_fun(pars, σ).real,
            apply_fun,
        )
        return log_pdf

    def init_state(
        self,
        machine: ModuleOrApplyFun,
        parameters: PyTree,
        seed: SeedT | None = None,
    ) -> SamplerState:
        """
        Creates the structure holding the state of the sampler.

        If you want reproducible samples, you should specify `seed`, otherwise the state
        will be initialised randomly.

        If running across several MPI processes, all `sampler_state`s are guaranteed to be
        in a different (but deterministic) state.
        This is achieved by first reducing (summing) the seed provided to every MPI rank,
        then generating `n_rank` seeds starting from the reduced one, and every rank is
        initialized with one of those seeds.

        The resulting state is guaranteed to be a frozen Python dataclass (in particular,
        a Flax dataclass), and it can be serialized using Flax serialization methods.

        Args:
            machine: A Flax module or callable with the forward pass of the log-pdf.
                If it is a callable, it should have the signature :code:`f(parameters, σ) -> jax.Array`.
            parameters: The PyTree of parameters of the model.
            seed: An optional seed or jax PRNGKey. If not specified, a random seed will be used.

        Returns:
            The structure holding the state of the sampler. In general you should not expect
            it to be in a valid state, and should reset it before use.
        """
        key = nkjax.PRNGKey(seed)
        key = nkjax.mpi_split(key)

        return self._init_state(wrap_afun(machine), parameters, key)

    def reset(
        self,
        machine: ModuleOrApplyFun,
        parameters: PyTree,
        state: SamplerState | None = None,
    ) -> SamplerState:
        """
        Resets the state of the sampler. To be used every time the parameters are changed.

        Args:
            machine: A Flax module or callable with the forward pass of the log-pdf.
                If it is a callable, it should have the signature :code:`f(parameters, σ) -> jax.Array`.
            parameters: The PyTree of parameters of the model.
            state: The current state of the sampler. If not specified, it will be constructed
                by calling :code:`sampler.init_state(machine, parameters)` with a random seed.

        Returns:
            A valid sampler state.
        """
        if state is None:
            state = self.init_state(machine, parameters)

        return self._reset(wrap_afun(machine), parameters, state)

    @overload
    def sample(
        self,
        machine: ModuleOrApplyFun,
        parameters: PyTree,
        *,
        state: SamplerState | None = None,
        chain_length: int = 1,
        return_log_probabilities: Literal[False] = False,
    ) -> tuple[jax.Array, SamplerState]: ...

    @overload
    def sample(
        self,
        machine: Callable | nn.Module,
        parameters: PyTree,
        *,
        state: SamplerState | None = None,
        chain_length: int = 1,
        return_log_probabilities: Literal[True],
    ) -> tuple[tuple[jax.Array, jax.Array], SamplerState]: ...

    def sample(
        self,
        machine: Callable | nn.Module,
        parameters: PyTree,
        *,
        state: SamplerState | None = None,
        chain_length: int = 1,
        return_log_probabilities: bool = False,
    ) -> (
        tuple[jax.Array, SamplerState]
        | tuple[tuple[jax.Array, jax.Array], SamplerState]
    ):
        """
        Samples `chain_length` batches of samples along the chains.

        Arguments:
            machine: A Flax module or callable with the forward pass of the log-pdf.
                If it is a callable, it should have the signature :code:`f(parameters, σ) -> jax.Array`.
            parameters: The PyTree of parameters of the model.
            state: The current state of the sampler. If not specified, then initialize and reset it.
            chain_length: The length of the chains (default = 1).
            return_log_probabilities: If `True`, the log-probabilities are also returned, which is sometimes
                useful to avoid re-evaluating the log-pdf when doing importance sampling. Defaults to False.

        Returns:
            Returns a tuple of 'samples' and 'state'. If `return_log_probabilities` is False,
            the samples are just the 3-rank array of samples. If `return_log_probabilities` is
            True, the samples are a tuple of the 3-rank array of samples and the 2-rank array of
            un-normalized log-probabilities corresponding to each sample.
        """
        if state is None:
            state = self.reset(machine, parameters)

        return self._sample_chain(
            wrap_afun(machine),
            parameters,
            state,
            chain_length,
            return_log_probabilities=return_log_probabilities,
        )

    def samples(
        self,
        machine: ModuleOrApplyFun,
        parameters: PyTree,
        *,
        state: SamplerState | None = None,
        chain_length: int = 1,
    ) -> Iterator[jax.Array]:
        """
        Returns a generator sampling `chain_length` batches of samples along the chains.

        Arguments:
            machine: A Flax module or callable with the forward pass of the log-pdf.
                If it is a callable, it should have the signature :code:`f(parameters, σ) -> jax.Array`.
            parameters: The PyTree of parameters of the model.
            state: The current state of the sampler. If not specified, then initialize and reset it.
            chain_length: The length of the chains (default = 1).
        """
        if state is None:
            state = self.reset(machine, parameters)

        machine = wrap_afun(machine)

        for _i in range(chain_length):
            samples, state = self._sample_chain(machine, parameters, state, 1)
            yield samples[:, 0, :]

    @abc.abstractmethod
    def _sample_chain(
        self,
        machine: nn.Module,
        parameters: PyTree,
        state: SamplerState,
        chain_length: int,
        return_log_probabilities: bool = False,
    ) -> (
        tuple[jax.Array, SamplerState]
        | tuple[tuple[jax.Array, jax.Array], SamplerState]
    ):
        """
        Implementation of `sample` for subclasses of `Sampler`.

        If you subclass `Sampler`, you should override this and not `sample`
        itself, because `sample` contains some common logic.

        If using Jax, this function should be jitted.

        Arguments:
            machine: A Flax module with the forward pass of the log-pdf.
            parameters: The PyTree of parameters of the model.
            state: The current state of the sampler.
            chain_length: The length of the chains.
            return_log_probabilities: If `True`, the log-probabilities are also returned.
                Defaults to False.

        Returns:
            Returns a tuple of 'samples' and 'state'. If `return_log_probabilities` is False,
            the samples are just the 3-rank array of samples. If `return_log_probabilities` is
            True, the samples are a tuple of the 3-rank array of samples and the 2-rank array of
            un-normalized log-probabilities corresponding to each sample.
        """

    @abc.abstractmethod
    def _init_state(self, machine, params, seed) -> SamplerState:
        """
        Implementation of `init_state` for subclasses of `Sampler`.

        If you subclass `Sampler`, you should override this and not `init_state`
        itself, because `init_state` contains some common logic.
        """

    @abc.abstractmethod
    def _reset(self, machine, parameters, state):
        """
        Implementation of `reset` for subclasses of `Sampler`.

        If you subclass `Sampler`, you should override this and not `reset`
        itself, because `reset` contains some common logic.
        """



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import numpy as np

from jax import numpy as jnp

from netket.graph import AbstractGraph

from .base import MetropolisRule


class ExchangeRule(MetropolisRule):
    r"""
    A Rule exchanging the state on a random couple of sites, chosen from a list of
    possible couples (clusters).

    This rule acts on two local degree of freedom :math:`s_i` and :math:`s_j`,
    and proposes a new state: :math:`s_1 \dots s^\prime_i \dots s^\prime_j \dots s_N`,
    where in general :math:`s^\prime_i \neq s_i` and :math:`s^\prime_j \neq s_j`.
    The sites :math:`i` and :math:`j` are also chosen to be within a maximum graph
    distance of :math:`d_{\mathrm{max}}`.

    The transition probability associated to this sampler can
    be decomposed into two steps:

    1. A pair of indices :math:`i,j = 1\dots N`, and such
       that :math:`\mathrm{dist}(i,j) \leq d_{\mathrm{max}}`,
       is chosen with uniform probability, excluding all the sites where
       :math:`s_i = s_j`;
    2. The sites are exchanged, i.e. :math:`s^\prime_i = s_j` and :math:`s^\prime_j = s_i`;

    Notice that this sampling method generates random permutations of the quantum
    numbers, thus global quantities such as the sum of the local quantum numbers
    are conserved during the sampling.
    This scheme should be used then only when sampling in a
    region where :math:`\sum_i s_i = \mathrm{constant}` is needed,
    otherwise the sampling would be strongly not ergodic.

    .. warning::

        If you are working with systems where the number of nodes in the physical lattice
        does not match the number of degrees of freedom, you must be careful!

        A typical example is a system of Spin-1/2 fermions on a lattice with N sites, where the
        first N degrees of freedom correspond to the spin down degrees of freedom and the
        next N degrees of freedom correspond to the spin up degrees of freedom.

        In this case, you tipically want to exchange only degrees of freedom of the same type.
        A simple way to achieve this is to double the graph:

        .. code-block:: python

            import netket as nk
            g = nk.graph.Square(5)
            hi = nk.hilbert.SpinOrbitalFermions(g.n_nodes, s=0.5)

            exchange_graph = nk.graph.disjoint_union(g, g)
            print("Exchange graph size:", exchange_graph.n_nodes)

            sa = nk.sampler.MetropolisExchange(hi, graph=exchange_graph, d_max=1)

    """

    clusters: jax.Array
    r"""2-Dimensional tensor :math:`T_{i,j}` of shape
    :math:`N_\text{clusters}\times 2` where the first dimension
    runs over the list of 2-site clusters and the second dimension
    runs over the 2 sites of those clusters.

    The Exchange rule will swap the two sites of a random row of this
    matrix at every Metropolis step.
    """

    def __init__(
        self,
        *,
        clusters: list[tuple[int, int]] | None = None,
        graph: AbstractGraph | None = None,
        d_max: int = 1,
    ):
        r"""
        Constructs the Exchange Rule.

        You can pass either a list of clusters or a netket graph object to
        determine the clusters to exchange.

        Args:
            clusters: The list of clusters that can be exchanged. This should be
                a list of 2-tuples containing two integers. Every tuple is an edge,
                or cluster of sites to be exchanged.
            graph: A graph, from which the edges determine the clusters
                that can be exchanged.
            d_max: Only valid if a graph is passed in. The maximum distance
                between two sites
        """
        if clusters is None and graph is not None:
            clusters = compute_clusters(graph, d_max)
        elif not (clusters is not None and graph is None):
            raise ValueError(
                """You must either provide the list of exchange-clusters or a netket graph, from
                              which clusters will be computed using the maximum distance d_max. """
            )

        self.clusters = jnp.array(clusters)

    def transition(rule, sampler, machine, parameters, state, key, σ):
        n_chains = σ.shape[0]

        # compute a mask for the clusters that can be hopped
        hoppable_clusters = _compute_different_clusters_mask(rule.clusters, σ)

        keys = jnp.asarray(jax.random.split(key, n_chains))

        @jax.vmap
        def _update_samples(key, σ, hoppable_clusters):
            # pick a random cluster, taking into account the mask
            n_conn = hoppable_clusters.sum(axis=-1)
            cluster = jax.random.choice(
                key,
                a=jnp.arange(rule.clusters.shape[0]),
                p=hoppable_clusters,
                replace=True,
            )

            # sites to be exchanged
            si = rule.clusters[cluster, 0]
            sj = rule.clusters[cluster, 1]

            σp = σ.at[si].set(σ[sj])
            σp = σp.at[sj].set(σ[si])

            # compute the number of connected sites
            hoppable_clusters_proposed = _compute_different_clusters_mask(
                rule.clusters, σp
            )
            n_conn_proposed = hoppable_clusters_proposed.sum(axis=-1)
            log_prob_corr = jnp.log(n_conn) - jnp.log(n_conn_proposed)
            return σp, log_prob_corr

        return _update_samples(keys, σ, hoppable_clusters)

    def __repr__(self):
        return f"ExchangeRule(# of clusters: {len(self.clusters)})"


def compute_clusters(graph: AbstractGraph, d_max: int):
    """
    Given a netket graph and a maximum distance, computes all clusters.
    If `d_max = 1` this is equivalent to taking the edges of the graph.
    Then adds next-nearest neighbors and so on.
    """
    clusters = []
    distances = np.asarray(graph.distances())
    size = distances.shape[0]
    for i in range(size):
        for j in range(i + 1, size):
            if distances[i][j] <= d_max:
                clusters.append((i, j))

    res_clusters = np.empty((len(clusters), 2), dtype=np.int64)

    for i, cluster in enumerate(clusters):
        res_clusters[i] = np.asarray(cluster)

    return res_clusters


@jax.jit
def _compute_different_clusters_mask(clusters, σ):
    # mask the clusters to include only moves
    # where the dof changes
    if jnp.issubdtype(σ, jnp.bool) or jnp.issubdtype(σ, jnp.integer):
        hoppable_clusters_mask = σ[..., clusters[:, 0]] != σ[..., clusters[:, 1]]
    else:
        hoppable_clusters_mask = ~jnp.isclose(
            σ[..., clusters[:, 0]], σ[..., clusters[:, 1]]
        )
    return hoppable_clusters_mask



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import solver
from . import qgt

from .linear_operator import LinearOperator
from .preconditioner import (
    LinearPreconditioner,
    PreconditionerT,
    identity_preconditioner,
)

from .sr import SR

from netket.utils import _hide_submodules


## Optimisers


def Sgd(learning_rate: float):
    r"""Stochastic Gradient Descent Optimizer.
    The `Stochastic Gradient Descent <https://en.wikipedia.org/wiki/Stochastic_gradient_descent>`_
    is one of the most popular optimizers in machine learning applications.
    Given a stochastic estimate of the gradient of the cost function (:math:`G(\mathbf{p})`),
    it performs the update:

    .. math:: p^\prime_k = p_k -\eta G_k(\mathbf{p}),

    where :math:`\eta` is the so-called learning rate.
    NetKet also implements two extensions to the simple SGD,
    the first one is :math:`L_2` regularization,
    and the second one is the possibility to set a decay
    factor :math:`\gamma \leq 1` for the learning rate, such that
    at iteration :math:`n` the learning rate is :math:`\eta \gamma^n`.

    Args:
       learning_rate: The learning rate :math:`\eta`.

    Examples:
       Simple SGD optimizer.

       >>> from netket.optimizer import Sgd
       >>> op = Sgd(learning_rate=0.05)
    """
    from optax import sgd

    return sgd(learning_rate)


def Momentum(learning_rate: float, beta: float = 0.9, nesterov: bool = False):
    r"""Momentum-based Optimizer.
        The momentum update incorporates an exponentially weighted moving average
        over previous gradients to speed up descent
        `Qian, N. (1999) <http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.57.5612&rep=rep1&type=pdf>`_.
        The momentum vector :math:`\mathbf{m}` is initialized to zero.
        Given a stochastic estimate of the gradient of the cost function
        :math:`G(\mathbf{p})`, the updates for the parameter :math:`p_k` and
        corresponding component of the momentum :math:`m_k` are

        .. math::

            m^\prime_k &= \beta m_k + (1-\beta)G_k(\mathbf{p})\\
            p^\prime_k &= \eta m^\prime_

        Args:
           learning_rate: The learning rate :math:`\eta`
           beta: Momentum exponential decay rate, should be in [0,1].
           nesterov: Flag to use nesterov momentum correction

        Examples:
           Momentum optimizer.

           >>> from netket.optimizer import Momentum
           >>> op = Momentum(learning_rate=0.01)
    """
    from optax import sgd

    return sgd(learning_rate, momentum=beta, nesterov=nesterov)


def AdaGrad(
    learning_rate: float = 0.001,
    epscut: float = 1.0e-7,
    initial_accumulator_value: float = 0.1,
):
    r"""AdaGrad Optimizer.
        In many cases, in Sgd the learning rate :math:`\eta` should
        decay as a function of training iteration to prevent overshooting
        as the optimum is approached. AdaGrad is an adaptive learning
        rate algorithm that automatically scales the learning rate with a sum
        over past gradients. The vector :math:`\mathbf{g}` is initialized to zero.
        Given a stochastic estimate of the gradient of the cost function :math:`G(\mathbf{p})`,
        the updates for :math:`g_k` and the parameter :math:`p_k` are


        .. math:: g^\prime_k &= g_k + G_k(\mathbf{p})^2\\
                  p^\prime_k &= p_k - \frac{\eta}{\sqrt{g_k + \epsilon}}G_k(\mathbf{p})

        AdaGrad has been shown to perform particularly well when
        the gradients are sparse, but the learning rate may become too small
        after many updates because the sum over the squares of past gradients is cumulative.


        Args:
           learning_rate: Learning rate :math:`\eta`.
           epscut: Small :math:`\epsilon` cutoff.
           initial_accumulator_value: initial value of the accumulator

        Examples:
           Simple AdaGrad optimizer.

           >>> from netket.optimizer import AdaGrad
           >>> op = AdaGrad()
        """
    from optax import adagrad

    return adagrad(
        learning_rate, eps=epscut, initial_accumulator_value=initial_accumulator_value
    )


def Adam(learning_rate: float = 0.001, b1: float = 0.9, b2: float = 0.999, eps=1e-08):
    r"""Adam Optimizer.

    Args:
        learning_rate: Learning rate :math:`\eta`.
        b1: Decay rate for the exponentially weighted average of grads.
        b2: Decay rate for the exponentially weighted average of squared norm of grads.
        eps: Term added to the denominator to improve numerical stability.
    """
    from optax import adam

    return adam(learning_rate, b1=b1, b2=b2, eps=eps)


def RmsProp(
    learning_rate: float = 0.001,
    beta: float = 0.9,
    epscut: float = 1.0e-7,
    centered: bool = False,
):
    r"""RMSProp optimizer.

    RMSProp is a well-known update algorithm proposed by Geoff Hinton
    in his Neural Networks course notes `Neural Networks course notes
    <http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf>`_.
    It corrects the problem with AdaGrad by using an exponentially weighted
    moving average over past squared gradients instead of a cumulative sum.
    After initializing the vector :math:`\mathbf{s}` to zero, :math:`s_k` and t
    he parameters :math:`p_k` are updated as

    .. math:: s^\prime_k = \beta s_k + (1-\beta) G_k(\mathbf{p})^2 \\
              p^\prime_k = p_k - \frac{\eta}{\sqrt{s_k}+\epsilon} G_k(\mathbf{p})


    Constructs a new ``RmsProp`` optimizer.

    Args:
       learning_rate: The learning rate :math:`\eta`
       beta: Exponential decay rate.
       epscut: Small cutoff value.
       centered: whether to center the moving average.

    Examples:
       RmsProp optimizer.

       >>> from netket.optimizer import RmsProp
       >>> op = RmsProp(learning_rate=0.02)
    """
    from optax import rmsprop

    return rmsprop(learning_rate, decay=beta, eps=epscut, centered=centered)


_hide_submodules(__name__)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
import jax.numpy as jnp

from netket.utils import mpi


def to_shift_offset(
    diag_shift: float | None, diag_scale: float | None
) -> tuple[float, float | None]:
    if diag_shift is None:
        diag_shift = 0.0

    if isinstance(diag_scale, jax.Array):
        return diag_scale, diag_shift / diag_scale
    else:
        if diag_scale is None or diag_scale == 0.0:
            return diag_shift, None
        else:
            return diag_scale, diag_shift / diag_scale


@partial(jax.jit, static_argnames="ndims")
def rescale(centered_oks, offset, *, ndims: int = 1):
    """
    compute ΔOₖ/√Sₖₖ and √Sₖₖ
    to do scale-invariant regularization (Becca & Sorella 2017, pp. 143)
    Sₖₗ/(√Sₖₖ√Sₗₗ) = ΔOₖᴴΔOₗ/(√Sₖₖ√Sₗₗ) = (ΔOₖ/√Sₖₖ)ᴴ(ΔOₗ/√Sₗₗ)

    Args:
        centered_oks: A mean-zero Jacobian.
        ndims: A number of leading dimensions to use to compute the
            rescale factor. Those should be all the non-parameters
            axes in the jacobian (so it should be 1 normally, 2 for
            non holomorphic stacked jacobians).
    """
    # should be (0,) for standard, (0,1) when we have 2 jacobians in complex mode
    axis = tuple(range(ndims))

    scale = jax.tree_util.tree_map(
        lambda x: (
            mpi.mpi_sum_jax(jnp.sum((x * x.conj()).real, axis=axis, keepdims=True))[0]
            + offset
        )
        ** 0.5,
        centered_oks,
    )
    centered_oks = jax.tree_util.tree_map(jnp.divide, centered_oks, scale)
    scale = jax.tree_util.tree_map(partial(jnp.squeeze, axis=axis), scale)
    return centered_oks, scale



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tarfile as _tarfile
from os import path as _path

from flax import serialization as _serialization
from netket.utils.types import PyTree as _PyTree


def variables_from_file(filename: str, variables: _PyTree):
    """
    Loads the variables of a variational state from a `.mpack` file.

    Args:
        filename: the file containing the variables. Assumes a .mpack
            extension and adds it if missing and no file exists.
        variables: An object variables with the same structure and shape
            of the object to be deserialized.

    Returns:
        a PyTree like variables

    Examples:
       Serializing the data:

       >>> import netket as nk
       >>> import flax
       >>> # construct an RBM model on 10 spins
       >>> vstate = nk.vqs.MCState(
       ...      nk.sampler.MetropolisLocal(nk.hilbert.Spin(0.5)**10),
       ...      nk.models.RBM())
       >>> with open("test.mpack", 'wb') as file:
       ...     bytes_written = file.write(flax.serialization.to_bytes(vstate.variables))
       >>> print(bytes_written)
       1052
       >>>
       >>> # Deserialize the data
       >>>
       >>> del vstate
       >>> # construct an RBM model on 10 spins
       >>> vstate2 = nk.vqs.MCState(
       ...      nk.sampler.MetropolisLocal(nk.hilbert.Spin(0.5)**10),
       ...      nk.models.RBM())
       >>> # Load the data by passing the model
       >>> vars = nk.experimental.vqs.variables_from_file("test.mpack",
       ...                                                        vstate2.variables)
       >>> # update the variables of vstate with the loaded data.
       >>> vstate2.variables = vars
    """
    if not _path.isfile(filename):
        if filename[-6:] != ".mpack":
            filename = filename + ".mpack"

    with open(filename, "rb") as f:
        return _serialization.from_bytes(variables, f.read())


def variables_from_tar(filename: str, variables: _PyTree, i: int):
    """
    Loads the variables of a variational state from the i-th element of a `.tar`
    archive.

    Args:
        filename: the tar archive name. Assumes a .tar
            extension and adds it if missing and no file exists.
        variables: An object variables with the same structure and shape
            of the object to be deserialized.
        i: the index of the variables to load
    """
    if not _path.isfile(filename):
        if filename[-4:] != ".tar":
            filename = filename + ".tar"

    with _tarfile.TarFile(filename, "r") as file:
        info = file.getmember(str(i) + ".mpack")
        with file.extractfile(info) as f:
            return _serialization.from_bytes(variables, f.read())



from functools import partial

import jax
import jax.numpy as jnp

from jax.api_util import argnums_partial

from jax.extend import linear_util as lu

_tree_add = partial(jax.tree_util.tree_map, jax.lax.add)
_tree_zeros_like = partial(
    jax.tree_util.tree_map, lambda x: jnp.zeros(x.shape, dtype=x.dtype)
)


# TODO put it somewhere
def _multimap(f, *args):
    try:
        return tuple(map(lambda a: f(*a), zip(*args)))
    except TypeError:
        return f(*args)


def scan_append_reduce(f, x, append_cond, op=_tree_add, zero_fun=_tree_zeros_like):
    """Evaluate f element by element in x while appending and/or reducing the results

    Args:
        f: a function that takes elements of the leading dimension of x
        x: a pytree where each leaf array has the same leading dimension
        append_cond: a bool (if f returns just one result) or a tuple of bools (if f returns multiple values)
            which indicates whether the individual result should be appended or reduced
        op: a function to (pairwise) reduce the specified results. Defaults to a sum.
        zero_fun: a function which prepares the zero element of op for a given input shape/dtype tree. Defaults to zeros.

    Returns:
        returns the (tuple of) results corresponding to the output of f
        where each result is given by:
        if append_cond is True:
            a (pytree of) array(s) with leading dimension same as x,
            containing the evaluation of f at each element in x
        else (append_cond is False):
            a (pytree of) array(s) with the same shape as the corresponding output of f,
            containing the reduction over op of f evaluated at each x


    Example:

        import jax.numpy as jnp
        from netket.jax import scan_append_reduce

        def f(x):
             y = jnp.sin(x)
             return y, y, y**2

        N = 100
        x = jnp.linspace(0.,jnp.pi,N)

        y, s, s2 = scan_append_reduce(f, x, (True, False, False))
        mean = s/N
        var = s2/N - mean**2
    """
    # TODO: different op for each result

    x0 = jax.tree_util.tree_map(lambda x: x[0], x)

    # special code path if there is only one element
    # to avoid having to rely on xla/llvm to optimize the overhead away
    if jax.tree_util.tree_leaves(x)[0].shape[0] == 1:
        return _multimap(
            lambda c, x: jnp.expand_dims(x, 0) if c else x, append_cond, f(x0)
        )

    # the original idea was to use pytrees, however for now just operate on the return value tuple
    _get_append_part = partial(_multimap, lambda c, x: x if c else None, append_cond)
    _get_op_part = partial(_multimap, lambda c, x: x if not c else None, append_cond)
    _tree_select = partial(_multimap, lambda c, t1, t2: t1 if c else t2, append_cond)

    carry_init = True, _get_op_part(zero_fun(jax.eval_shape(f, x0)))

    def f_(carry, x):
        is_first, y_carry = carry
        y = f(x)
        y_op = _get_op_part(y)
        y_append = _get_append_part(y)
        y_reduce = op(y_carry, y_op)
        return (False, y_reduce), y_append

    (_, res_op), res_append = jax.lax.scan(f_, carry_init, x, unroll=1)
    # reconstruct the result from the reduced and appended parts in the two trees
    return _tree_select(res_append, res_op)


scan_append = partial(scan_append_reduce, append_cond=True)
scan_reduce = partial(scan_append_reduce, append_cond=False)


# TODO in_axes a la vmap?
def scanmap(fun, scan_fun, argnums=0):
    """
    A helper function to wrap f with a scan_fun

    Example:
        import jax.numpy as jnp
        from functools import partial

        from netket.jax import scanmap, scan_append_reduce

        scan_fun = partial(scan_append_reduce, append_cond=(True, False, False))

        @partial(scanmap, scan_fun=scan_fun, argnums=1)
        def f(c, x):
             y = jnp.sin(x) + c
             return y, y, y**2

        N = 100
        x = jnp.linspace(0.,jnp.pi,N)
        c = 1.


        y, s, s2 = f(c, x)
        mean = s/N
        var = s2/N - mean**2
    """

    def f_(*args, **kwargs):
        f = lu.wrap_init(fun, kwargs)
        f_partial, dyn_args = argnums_partial(
            f, argnums, args, require_static_args_hashable=False
        )
        return scan_fun(lambda x: f_partial.call_wrapped(*x), dyn_args)

    return f_



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar
from collections.abc import Callable, Iterable

import abc
import numbers
from functools import partial

from tqdm.auto import tqdm

import jax

from netket import config
from netket.logging import AbstractLog, JsonLog
from netket.operator._abstract_observable import AbstractObservable
from netket.utils import mpi, timing
from netket.utils.types import Optimizer, PyTree
from netket.vqs import VariationalState

CallbackT = Callable[[int, dict, "AbstractVariationalDriver"], bool]

T = TypeVar("T")


def _to_iterable(maybe_iterable: T | Iterable[T]) -> tuple[T, ...]:
    """
    _to_iterable(maybe_iterable)

    Ensure the result is iterable. If the input is not iterable, it is wrapped into a tuple.
    """
    if not isinstance(maybe_iterable, Iterable):
        maybe_iterable = (maybe_iterable,)

    return tuple(maybe_iterable)


class AbstractVariationalDriver(abc.ABC):
    """Abstract base class for NetKet Variational Monte Carlo drivers

    This class must be inherited from in order to create an optimization driver that
    immediately works with NetKet loggers and callback mechanism.

    .. note::

        How to implement a new driver

        For a concrete example, look at the file `netket/driver/vmc.py`.

        If you want to inherit the nice interface of :class:`netket.driver.AbstractVariationalDriver`,
        you should subclass it, and define the following methods:

        - The :meth:`~netket.driver.AbstractVariationalDriver.__init__` method should be called
          with the machine, optimizer and optionally the name of the loss minimised. If this
          driver is minimising a loss function and you want it's name to show up automatically
          in the progress bar/output files you should pass the optional keyword argument.

        - :meth:`~netket.driver.AbstractVariationalDriver._forward_and_backward`,
          that should compute the loss function and the gradient, returning the latter.
          If the driver is minimizing or maximising some loss function,
          this quantity should be assigned to the field `self._loss_stats`
          in order to monitor it.

        - :meth:`~netket.driver.AbstractVariationalDriver._estimate_stats` should return
          the expectation value over the variational state of a single observable.

        - :meth:`~netket.driver.AbstractVariationalDriver.reset`,
          should reset the driver (usually the sampler). The basic implementation will call
          :meth:`~netket.vqs.VariationalState.reset`, but you are responsible for resetting
          extra fields in the driver itself.

    """

    def __init__(
        self,
        variational_state: VariationalState,
        optimizer: Optimizer,
        minimized_quantity_name: str = "loss",
    ):
        """
        Initializes a variational optimization driver.

        Args:
            variational_state: The variational state to be optimized
            optimizer: an `optax <https://optax.readthedocs.io/en/latest/>`_ optimizer.
                If you do not want
                to use an optimizer, just pass a sgd optimizer with
                learning rate `-1`.
            minimized_quantity_name: the name of the loss function in
                the logged data set.
        """
        self._mynode = mpi.node_number
        self._is_root = self._mynode == 0 and jax.process_index() == 0
        self._mpi_nodes = mpi.n_nodes
        self._loss_stats = None
        self._loss_name = minimized_quantity_name
        self._step_count = 0
        self._timer = None

        self._variational_state = variational_state
        self.optimizer = optimizer

    def _forward_and_backward(self) -> PyTree:  # pragma: no cover
        """
        :meta public:

        Performs a step of the optimization driver, returning the PyTree
        of the gradients that will be optimized.

        Concrete drivers must override this method.

        .. note::

            When implementing this function on a subclass, you must return the
            gradient which must match the pytree structure of the parameters
            of the variational state.

            The gradient will then be passed on to the optimizer in order to update
            the parameters.

            Moreover, if you are minimising a loss function you must set the
            field `self._loss_stats` with the current value of the loss function.

            This will be logged to any logger during optimisation.

        Returns:
            the update for the weights.
        """
        raise NotImplementedError()  # pragma: no cover

    def _estimate_stats(self, observable):
        """
        Returns the MCMC statistics for the expectation value of an observable.
        Must be implemented by super-classes of AbstractVMC.

        Args:
            observable: A quantum operator (netket observable)

        Returns:
            The expectation value of the observable.
        """
        return self.state.expect(observable)

    def _log_additional_data(self, log_dict: dict, step: int):
        """
        Method to be implemented in sub-classes of AbstractVariationalDriver to
        log additional data at every step.
        This method is called at every iteration when executing with `run`.

        Args:
            log_dict: The dictionary containing all logged data. It must be
                **modified in-place** adding new keys.
            step: the current step number.
        """
        # Always log the acceptance.
        if hasattr(self.state, "sampler_state"):
            acceptance = getattr(self.state.sampler_state, "acceptance", None)  # type: ignore
            if acceptance is not None:
                log_dict["acceptance"] = acceptance

    def reset(self):
        """
        Resets the driver.

        Subclasses should make sure to call :code:`super().reset()` to ensure
        that the step count is set to 0.
        """
        self.state.reset()
        self._step_count = 0

    @property
    def state(self):
        """
        Returns the machine that is optimized by this driver.
        """
        return self._variational_state

    @property
    def optimizer(self):
        """
        The optimizer used to update the parameters at every iteration.
        """
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer):
        self._optimizer = optimizer
        if optimizer is not None:
            self._optimizer_state = optimizer.init(self.state.parameters)
            if config.netket_experimental_sharding:
                self._optimizer_state = jax.lax.with_sharding_constraint(
                    self._optimizer_state,
                    jax.sharding.PositionalSharding(jax.devices()).replicate(),
                )

    @property
    def step_count(self):
        """
        Returns a monotonic integer labelling all the steps performed by this driver.
        This can be used, for example, to identify the line in a log file.
        """
        return self._step_count

    def iter(self, n_steps: int, step: int = 1):
        """
        Returns a generator which advances the VMC optimization, yielding
        after every `step_size` steps.

        Args:
            n_steps: The total number of steps to perform (this is
                equivalent to the length of the iterator)
            step: The number of internal steps the simulation
                is advanced between yielding from the iterator

        Yields:
            int: The current step.
        """
        for _ in range(0, n_steps, step):
            for i in range(0, step):
                self._dp = self._forward_and_backward()
                if i == 0:
                    yield self.step_count

                self._step_count += 1
                self.update_parameters(self._dp)

    def advance(self, steps: int = 1):
        """
        Performs `steps` optimization steps.

        Args:
            steps: (Default=1) number of steps.

        """
        for _ in self.iter(steps):
            pass

    def run(
        self,
        n_iter: int,
        out: AbstractLog | Iterable[AbstractLog] | str | None = (),
        obs: dict[str, AbstractObservable] | None = None,
        step_size: int = 1,
        show_progress: bool = True,
        save_params_every: int = 50,  # for default logger
        write_every: int = 50,  # for default logger
        callback: CallbackT | Iterable[CallbackT] = lambda *x: True,
        timeit: bool = False,
    ):
        """
        Runs this variational driver, updating the weights of the network stored in
        this driver for `n_iter` steps and dumping values of the observables `obs`
        in the output `logger`.

        It is possible to control more specifically what quantities are logged, when to
        stop the optimisation, or to execute arbitrary code at every step by specifying
        one or more callbacks, which are passed as a list of functions to the keyword
        argument `callback`.

        Callbacks are functions that follow this signature:

        .. Code::

            def callback(step, log_data, driver) -> bool:
                ...
                return True/False

        If a callback returns True, the optimisation continues, otherwise it is stopped.
        The `log_data` is a dictionary that can be modified in-place to change what is
        logged at every step. For example, this can be used to log additional quantities
        such as the acceptance rate of a sampler.

        Loggers are specified as an iterable passed to the keyword argument `out`. If only
        a string is specified, this will create by default a :class:`nk.logging.JsonLog`.
        To know about the output format check its documentation. The logger object is
        also returned at the end of this function so that you can inspect the results
        without reading the json output.

        When running among multiple MPI ranks/Jax devices, the logging logic is executed
        on all nodes, but only root-rank loggers should write to files or do expensive I/O
        operations.

        .. note::

            Before NetKet 3.15, loggers where automatically 'ignored' on non-root ranks.
            However, starting with NetKet 3.15 it is the responsability of a logger to
            check if it is executing on a non-root rank, and to 'do nothing' if that is
            the case.

            The change was required to work correctly and efficiently with sharding. It will
            only affect users that were defining custom loggers themselves.

        Args:
            n_iter: the total number of iterations to be performed during this run.
            out: A logger object, or an iterable of loggers, to be used to store simulation log and data.
                If this argument is a string, it will be used as output prefix for the standard JSON logger.
            obs: An iterable containing all observables that should be computed
            step_size: Every how many steps should observables be logged to disk (default=1)
            callback: Callable or list of callable callback functions to stop training given a condition
            show_progress: If true displays a progress bar (default=True)
            save_params_every: Every how many steps the parameters of the network should be
                serialized to disk (ignored if logger is provided)
            write_every: Every how many steps the json data should be flushed to disk (ignored if
                logger is provided)
            timeit: If True, provide timing information.
        """

        if not isinstance(n_iter, numbers.Number):
            raise ValueError(
                "n_iter, the first positional argument to `run`, must be a number!"
            )

        if obs is None:
            obs = {}

        # if out is a path, create an overwriting Json Log for output
        if isinstance(out, str):
            out = JsonLog(out, "w", save_params_every, write_every)
        elif out is None:
            out = ()

        loggers = _to_iterable(out)
        callbacks = _to_iterable(callback)
        callback_stop = False

        with timing.timed_scope(force=timeit) as timer:
            with tqdm(
                total=n_iter,
                disable=not show_progress or not self._is_root,
                dynamic_ncols=True,
            ) as pbar:
                old_step = self.step_count
                first_step = True

                for step in self.iter(n_iter, step_size):
                    log_data = self.estimate(obs)
                    self._log_additional_data(log_data, step)

                    # if the cost-function is defined then report it in the progress bar
                    if self._loss_stats is not None:
                        pbar.set_postfix_str(
                            self._loss_name + "=" + str(self._loss_stats)
                        )
                        log_data[self._loss_name] = self._loss_stats

                    # Execute callbacks before loggers because they can append to log_data
                    for callback in callbacks:
                        if not callback(step, log_data, self):
                            callback_stop = True

                    with timing.timed_scope(name="loggers"):
                        for logger in loggers:
                            logger(self.step_count, log_data, self.state)

                    if len(callbacks) > 0:
                        if mpi.mpi_any(callback_stop):
                            break

                    # Reset the timing of tqdm after the first step, to ignore compilation time
                    if first_step:
                        first_step = False
                        pbar.unpause()

                    # Update the progress bar
                    pbar.update(self.step_count - old_step)
                    old_step = self.step_count

                # Final update so that it shows up filled.
                pbar.update(self.step_count - old_step)

        # flush at the end of the evolution so that final values are saved to
        # file
        for logger in loggers:
            logger.flush(self.state)

        if timeit:
            self._timer = timer
            if self._is_root:
                print(timer)

        return loggers

    @timing.timed(name="estimate observables")
    def estimate(self, observables):
        """
        Return MCMC statistics for the expectation value of observables in the
        current state of the driver.

        Args:
            observables: A pytree of operators for which statistics should be computed.

        Returns:
            A pytree of the same structure as the input, containing MCMC statistics
            for the corresponding operators as leaves.
        """

        # Do not unpack operators, even if they are pytrees!
        # this is necessary to support jax operators.
        return jax.tree_util.tree_map(
            self._estimate_stats,
            observables,
            is_leaf=lambda x: isinstance(x, AbstractObservable),
        )

    def update_parameters(self, dp):
        """
        Updates the parameters of the machine using the optimizer in this driver

        Args:
            dp: the pytree containing the updates to the parameters
        """
        self._optimizer_state, self.state.parameters = apply_gradient(
            self._optimizer.update, self._optimizer_state, dp, self.state.parameters
        )


@partial(jax.jit, static_argnums=0)
def apply_gradient(optimizer_fun, optimizer_state, dp, params):
    import optax

    updates, new_optimizer_state = optimizer_fun(dp, optimizer_state, params)

    new_params = optax.apply_updates(params, updates)

    if config.netket_experimental_sharding:
        sharding = jax.sharding.PositionalSharding(jax.devices()).replicate()
        new_optimizer_state = jax.lax.with_sharding_constraint(
            new_optimizer_state, sharding
        )
        new_params = jax.lax.with_sharding_constraint(new_params, sharding)

    return new_optimizer_state, new_params
